from __future__ import annotations

import ctypes
import datetime
import json
import logging
import multiprocessing
import os
import tempfile
import time
import unittest
from multiprocessing import Process
from typing import Any, ClassVar

import httpx
import yara_x
from azul_bedrock import models_network as azm

from azul_runner import (
    DATA_HASH,
    FV,
    Event,
    Feature,
    FeatureValue,
    Job,
    JobResult,
    State,
    StorageProxyFile,
    add_settings,
    coordinator,
    local,
    monitor,
    network,
    settings,
)
from azul_runner.models import TaskExitCodeEnum

from . import mock_dispatcher as md
from . import plugin_support as sup


def dump(x):
    return json.loads(x.model_dump_json(exclude_defaults=True))


class TestBasePluginLive(unittest.TestCase):
    """
    Test cases for base plugin class - cases that talk to the mock server
    """

    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server, suitable for passing to a plugin's config['server']
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()
        # Dummy shared memory queue and ctype
        cls.dummy_queue: multiprocessing.Queue = multiprocessing.Queue()

        cls.editor = md.Editor(cls.server)
        cls.editor.set_stream(
            "b4b389c849d799d9331d5937cde7f0dfd297d76083242366cbef53b498cd6051", 200, b"small content"
        )

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()
        cls.dummy_queue.close()

    # ############### #
    # #### Tests #### #
    # ############### #

    def _wait_for_is_alive_to_be_false(self, plugin_sub: Process, max_wait_sec: int = 20):
        """Sleep up to max_wait_sec waiting for a plugin subprocess to no longer be alive."""
        total_sleep = 0
        while plugin_sub.is_alive() and total_sleep < max_wait_sec:
            time.sleep(1)
            total_sleep += 1

    def _del(self, event):
        del (
            event["timestamp"],
            event["entity"]["config"]["server"],
            event["entity"]["config"]["events_url"],
            event["entity"]["config"]["data_url"],
        )

    def get_last_req_params(self) -> dict:
        r = httpx.get("%s/mock/get_var/last_req_params" % self.server)
        r.raise_for_status()
        return r.json()

    def test_config_secret(self) -> None:
        p = sup.DummyPlugin(
            config={
                "events_url": self.server + "/depth_1",
                "password": "yolo",
                "secret_password": "pog",
                "data_url": self.server,
            }
        )
        net = network.Network(p)
        net.post_registrations()
        self.assertEqual(p.cfg.password, "yolo")
        self.assertEqual(p.cfg.secret_password, "pog")
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        self._del(out_event)

        self.assertIn("password", out_event["entity"]["config"])
        self.assertNotIn("secret_password", out_event["entity"]["config"])

    class DPRegistartionConfigOverride(sup.DummyPlugin):
        """Hello world."""

        FEATURES = [
            Feature(name="feat1", desc="", type=azm.FeatureType.String),
            Feature(name="per_stream_feat", desc="", type=azm.FeatureType.String),
        ]

    def test_registration_config_overrides(self):

        p = self.DPRegistartionConfigOverride(
            config={
                "server": self.server + "/depth_1",
                "name_suffix": "ASuffix",
                "name_remove_prefix": "DiffPrefix",
                "version_suffix": "Beta5",
                "security_override": "Alpha Beta",
                "filter_min_content_size": "100kib",  # ensure that bytesize is encoded ok
            }
        )
        net = network.Network(p)
        net.post_registrations()
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        self._del(out_event)
        print(out_event)
        self.assertEqual(
            out_event,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "author": {
                    "category": "plugin",
                    "name": "DPRegistartionConfigOverride-ASuffix",
                    "version": "1.0-Beta5",
                    "security": "Alpha Beta",
                },
                "entity": {
                    "category": "plugin",
                    "name": "DPRegistartionConfigOverride-ASuffix",
                    "version": "1.0-Beta5",
                    "security": "Alpha Beta",
                    "description": "Hello world.",
                    "features": [
                        {"name": "example_date", "desc": "Example datetime feature", "type": "datetime"},
                        {"name": "example_int", "desc": "Example int feature", "type": "integer"},
                        {"name": "example_path", "desc": "Example Filepath feature", "type": "filepath"},
                        {"name": "example_raw", "desc": "Example raw bytes feature", "type": "binary"},
                        {"name": "example_string", "desc": "Example string feature", "type": "string"},
                        {"name": "example_unspec", "desc": "Example feature of unspecified type", "type": "string"},
                        {"name": "example_uri", "desc": "Example URI feature", "type": "uri"},
                        {"name": "feat1", "desc": "", "type": "string"},
                        {
                            "name": "file_extension",
                            "desc": "File extension of the 'content' stream.",
                            "type": "string",
                        },
                        {
                            "name": "file_format",
                            "desc": "Assemblyline file type of the 'content' stream.",
                            "type": "string",
                        },
                        {"name": "filename", "desc": "Name on disk of the 'content' stream.", "type": "filepath"},
                        {"name": "magic", "desc": "File magic found for the 'content' stream.", "type": "string"},
                        {"name": "malformed", "desc": "File is malformed in some way.", "type": "string"},
                        {"name": "mime", "desc": "Mimetype found for the 'content' stream.", "type": "string"},
                        {"name": "per_stream_feat", "desc": "", "type": "string"},
                    ],
                    "config": {
                        "assume_streams_available": "false",
                        "concurrent_plugin_instances": "1",
                        "content_meta_cache_limit": "0",
                        "cur_mem_file_path": '"/sys/fs/cgroup/memory.current"',
                        "cur_mem_summary_file_path": '"/sys/fs/cgroup/memory.stat"',
                        "deployment_key": '""',
                        "enable_mem_limits": "false",
                        "filter_allow_event_types": "[]",
                        "filter_data_types": "{}",
                        "filter_max_content_size": "209715200",
                        "filter_min_content_size": "102400",
                        "filter_self": "false",
                        "heartbeat_interval": "30",
                        "max_mem_file_path": '"/sys/fs/cgroup/memory.max"',
                        "max_timeouts_before_exit": "100",
                        "max_value_length": "4000",
                        "max_values_per_feature": "1000",
                        "mem_poll_frequency_milliseconds": "1000",
                        "name_remove_prefix": '"DiffPrefix"',
                        "name_suffix": '"ASuffix"',
                        "not_ready_backoff": "5",
                        "plugin_depth_limit": "10",
                        "request_retry_count": "0",
                        "request_timeout": "15",
                        "require_expedite": "true",
                        "require_historic": "true",
                        "require_live": "true",
                        "run_timeout": "600",
                        "security_override": '"Alpha Beta"',
                        "test_input_data": "{}",
                        "used_mem_force_exit_frac": "0.9",
                        "used_mem_warning_frac": "0.8",
                        "version_suffix": '"Beta5"',
                        "watch_path": '""',
                        "watch_type": '""',
                        "watch_wait": "10",
                    },
                },
            },
        )

    def test_registration_inheritance(self):
        p = sup.DummyPluginFeatureInheritance(config={"events_url": self.server + "/depth_1", "data_url": self.server})
        net = network.Network(p)
        net.post_registrations()
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        self._del(out_event)
        print(out_event)
        self.assertEqual(
            out_event,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "author": {"category": "plugin", "name": "DummyPluginFeatureInheritance", "version": "2.0"},
                "entity": {
                    "category": "plugin",
                    "name": "DummyPluginFeatureInheritance",
                    "version": "2.0",
                    "description": "Test class to ensure that features accumulate between template plugins and their descendants.",
                    "features": [
                        {
                            "name": "descendant feature",
                            "desc": "A feature added by the child plugin",
                            "type": "string",
                        },
                        {"name": "example_date", "desc": "Example datetime feature", "type": "datetime"},
                        {"name": "example_int", "desc": "Example int feature", "type": "integer"},
                        {"name": "example_path", "desc": "Example Filepath feature", "type": "filepath"},
                        {"name": "example_raw", "desc": "Example raw bytes feature", "type": "binary"},
                        {"name": "example_string", "desc": "Example string feature", "type": "string"},
                        {"name": "example_unspec", "desc": "Child class redefining feature", "type": "string"},
                        {"name": "example_uri", "desc": "Example URI feature", "type": "uri"},
                        {
                            "name": "file_extension",
                            "desc": "File extension of the 'content' stream.",
                            "type": "string",
                        },
                        {
                            "name": "file_format",
                            "desc": "Assemblyline file type of the 'content' stream.",
                            "type": "string",
                        },
                        {"name": "filename", "desc": "Name on disk of the 'content' stream.", "type": "filepath"},
                        {"name": "magic", "desc": "File magic found for the 'content' stream.", "type": "string"},
                        {"name": "malformed", "desc": "File is malformed in some way.", "type": "string"},
                        {"name": "mime", "desc": "Mimetype found for the 'content' stream.", "type": "string"},
                    ],
                    "config": {
                        "assume_streams_available": "false",
                        "concurrent_plugin_instances": "1",
                        "content_meta_cache_limit": "0",
                        "cur_mem_file_path": '"/sys/fs/cgroup/memory.current"',
                        "cur_mem_summary_file_path": '"/sys/fs/cgroup/memory.stat"',
                        "deployment_key": '""',
                        "enable_mem_limits": "false",
                        "filter_allow_event_types": "[]",
                        "filter_data_types": "{}",
                        "filter_max_content_size": "209715200",
                        "filter_min_content_size": "0",
                        "filter_self": "false",
                        "heartbeat_interval": "30",
                        "max_mem_file_path": '"/sys/fs/cgroup/memory.max"',
                        "max_timeouts_before_exit": "100",
                        "max_value_length": "4000",
                        "max_values_per_feature": "1000",
                        "mem_poll_frequency_milliseconds": "1000",
                        "name_remove_prefix": '"AzulPlugin"',
                        "name_suffix": '""',
                        "not_ready_backoff": "5",
                        "plugin_depth_limit": "10",
                        "request_retry_count": "0",
                        "request_timeout": "15",
                        "require_expedite": "true",
                        "require_historic": "true",
                        "require_live": "true",
                        "run_timeout": "600",
                        "security_override": '""',
                        "test_input_data": "{}",
                        "used_mem_force_exit_frac": "0.9",
                        "used_mem_warning_frac": "0.8",
                        "version_suffix": '""',
                        "watch_path": '""',
                        "watch_type": '""',
                        "watch_wait": "10",
                    },
                },
            },
        )

    class DPRegistrationSecurityField(sup.DummyPlugin):
        """Test case plugin class for security field."""

        SECURITY = "security_1 other_security"

    def test_registration_security_field(self):
        p = self.DPRegistrationSecurityField(config={"events_url": self.server + "/depth_1", "data_url": self.server})
        net = network.Network(p)
        net.post_registrations()
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        self._del(out_event)
        print(out_event)
        self.assertEqual(
            out_event,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "author": {
                    "category": "plugin",
                    "name": "DPRegistrationSecurityField",
                    "version": "1.0",
                    "security": "security_1 other_security",
                },
                "entity": {
                    "category": "plugin",
                    "name": "DPRegistrationSecurityField",
                    "version": "1.0",
                    "security": "security_1 other_security",
                    "description": "Test case plugin class for security field.",
                    "features": [
                        {"name": "example_date", "desc": "Example datetime feature", "type": "datetime"},
                        {"name": "example_int", "desc": "Example int feature", "type": "integer"},
                        {"name": "example_path", "desc": "Example Filepath feature", "type": "filepath"},
                        {"name": "example_raw", "desc": "Example raw bytes feature", "type": "binary"},
                        {"name": "example_string", "desc": "Example string feature", "type": "string"},
                        {"name": "example_unspec", "desc": "Example feature of unspecified type", "type": "string"},
                        {"name": "example_uri", "desc": "Example URI feature", "type": "uri"},
                        {
                            "name": "file_extension",
                            "desc": "File extension of the 'content' stream.",
                            "type": "string",
                        },
                        {
                            "name": "file_format",
                            "desc": "Assemblyline file type of the 'content' stream.",
                            "type": "string",
                        },
                        {"name": "filename", "desc": "Name on disk of the 'content' stream.", "type": "filepath"},
                        {"name": "magic", "desc": "File magic found for the 'content' stream.", "type": "string"},
                        {"name": "malformed", "desc": "File is malformed in some way.", "type": "string"},
                        {"name": "mime", "desc": "Mimetype found for the 'content' stream.", "type": "string"},
                    ],
                    "config": {
                        "assume_streams_available": "false",
                        "concurrent_plugin_instances": "1",
                        "content_meta_cache_limit": "0",
                        "cur_mem_file_path": '"/sys/fs/cgroup/memory.current"',
                        "cur_mem_summary_file_path": '"/sys/fs/cgroup/memory.stat"',
                        "deployment_key": '""',
                        "enable_mem_limits": "false",
                        "filter_allow_event_types": "[]",
                        "filter_data_types": "{}",
                        "filter_max_content_size": "209715200",
                        "filter_min_content_size": "0",
                        "filter_self": "false",
                        "heartbeat_interval": "30",
                        "max_mem_file_path": '"/sys/fs/cgroup/memory.max"',
                        "max_timeouts_before_exit": "100",
                        "max_value_length": "4000",
                        "max_values_per_feature": "1000",
                        "mem_poll_frequency_milliseconds": "1000",
                        "name_remove_prefix": '"AzulPlugin"',
                        "name_suffix": '""',
                        "not_ready_backoff": "5",
                        "plugin_depth_limit": "10",
                        "request_retry_count": "0",
                        "request_timeout": "15",
                        "require_expedite": "true",
                        "require_historic": "true",
                        "require_live": "true",
                        "run_timeout": "600",
                        "security_override": '""',
                        "test_input_data": "{}",
                        "used_mem_force_exit_frac": "0.9",
                        "used_mem_warning_frac": "0.8",
                        "version_suffix": '""',
                        "watch_path": '""',
                        "watch_type": '""',
                        "watch_wait": "10",
                    },
                },
            },
        )

    class DummyPluginSecurityDict(sup.DummyPlugin):
        """Test case plugin class for security field dict."""

        SECURITY = "security_1 inc_grp TLP:WHITE"

    def test_registration_security_dict(self):
        """Test registration using a security dict instead of list."""

        p = self.DummyPluginSecurityDict(
            config={"events_url": self.server + "/depth_1", "deployment_key": "apple", "data_url": self.server}
        )
        net = network.Network(p)
        net.post_registrations()
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        self._del(out_event)
        print(out_event)
        self.assertEqual(
            out_event,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "author": {
                    "category": "plugin",
                    "name": "DummyPluginSecurityDict",
                    "version": "1.0",
                    "security": "security_1 inc_grp TLP:WHITE",
                },
                "entity": {
                    "category": "plugin",
                    "name": "DummyPluginSecurityDict",
                    "version": "1.0",
                    "security": "security_1 inc_grp TLP:WHITE",
                    "description": "Test case plugin class for security field dict.",
                    "features": [
                        {"name": "example_date", "desc": "Example datetime feature", "type": "datetime"},
                        {"name": "example_int", "desc": "Example int feature", "type": "integer"},
                        {"name": "example_path", "desc": "Example Filepath feature", "type": "filepath"},
                        {"name": "example_raw", "desc": "Example raw bytes feature", "type": "binary"},
                        {"name": "example_string", "desc": "Example string feature", "type": "string"},
                        {"name": "example_unspec", "desc": "Example feature of unspecified type", "type": "string"},
                        {"name": "example_uri", "desc": "Example URI feature", "type": "uri"},
                        {
                            "name": "file_extension",
                            "desc": "File extension of the 'content' stream.",
                            "type": "string",
                        },
                        {
                            "name": "file_format",
                            "desc": "Assemblyline file type of the 'content' stream.",
                            "type": "string",
                        },
                        {"name": "filename", "desc": "Name on disk of the 'content' stream.", "type": "filepath"},
                        {"name": "magic", "desc": "File magic found for the 'content' stream.", "type": "string"},
                        {"name": "malformed", "desc": "File is malformed in some way.", "type": "string"},
                        {"name": "mime", "desc": "Mimetype found for the 'content' stream.", "type": "string"},
                    ],
                    "config": {
                        "assume_streams_available": "false",
                        "concurrent_plugin_instances": "1",
                        "content_meta_cache_limit": "0",
                        "cur_mem_file_path": '"/sys/fs/cgroup/memory.current"',
                        "cur_mem_summary_file_path": '"/sys/fs/cgroup/memory.stat"',
                        "deployment_key": '"apple"',
                        "enable_mem_limits": "false",
                        "filter_allow_event_types": "[]",
                        "filter_data_types": "{}",
                        "filter_max_content_size": "209715200",
                        "filter_min_content_size": "0",
                        "filter_self": "false",
                        "heartbeat_interval": "30",
                        "max_mem_file_path": '"/sys/fs/cgroup/memory.max"',
                        "max_timeouts_before_exit": "100",
                        "max_value_length": "4000",
                        "max_values_per_feature": "1000",
                        "mem_poll_frequency_milliseconds": "1000",
                        "name_remove_prefix": '"AzulPlugin"',
                        "name_suffix": '""',
                        "not_ready_backoff": "5",
                        "plugin_depth_limit": "10",
                        "request_retry_count": "0",
                        "request_timeout": "15",
                        "require_expedite": "true",
                        "require_historic": "true",
                        "require_live": "true",
                        "run_timeout": "600",
                        "security_override": '""',
                        "test_input_data": "{}",
                        "used_mem_force_exit_frac": "0.9",
                        "used_mem_warning_frac": "0.8",
                        "version_suffix": '""',
                        "watch_path": '""',
                        "watch_type": '""',
                        "watch_wait": "10",
                    },
                },
            },
        )

    def test_fetch_job(self):
        net = network.Network(
            sup.DummyPlugin(
                config={"events_url": self.server + "/depth_1", "deployment_key": "apple", "data_url": self.server}
            )
        )
        r = net.fetch_job()
        self.assertTrue(isinstance(r, azm.BinaryEvent))

        net = network.Network(
            sup.DummyPlugin(
                config={"events_url": self.server + "/null", "deployment_key": "apple", "data_url": self.server}
            )
        )
        r = net.fetch_job()
        expected = dump(
            azm.BinaryEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="test-dummy",
                dequeued="test-dummy-dequeued",
                action=azm.BinaryAction.Enriched,
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                source=azm.Source(
                    name="source",
                    path=[],
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                ),
                author=azm.Author(name="TestServer", category="blah"),
                entity=azm.BinaryEvent.Entity(sha256="1234", datastreams=[], features=[]),
            )
        )
        self.assertEqual(dump(r), expected)
        # check query parameters
        print(self.get_last_req_params())
        self.assertEqual(
            {
                "name": ["DummyPlugin"],
                "version": ["1.0"],
                "deployment_key": ["apple"],
                "r-expedite": ["true"],
                "r-live": ["true"],
                "r-historic": ["true"],
                "r-under-content-size": ["209715200"],
                "count": ["1"],
                "deadline": ["10"],
            },
            self.get_last_req_params(),
        )

        # test with more advanced query parameters
        class AzulPluginCustom(sup.DummyPlugin):
            pass

        net = network.Network(
            AzulPluginCustom(
                config={
                    "server": self.server + "/null",
                    "require_historic": False,
                    "filter_max_content_size": "1000kb",
                    "filter_min_content_size": "1kb",
                    "filter_allow_event_types": ["extracted"],
                    "filter_self": True,
                    "filter_data_types": {"*": ["executable/windows/dos"], "blob": ["archive/gzip", "archive/bzip2"]},
                    "deployment_key": "apple",
                }
            )
        )
        r = net.fetch_job()
        print(self.get_last_req_params())
        self.assertEqual(
            {
                "name": ["Custom"],
                "version": ["1.0"],
                "count": ["1"],
                "deployment_key": ["apple"],
                "deadline": ["10"],
                "r-expedite": ["true"],
                "r-live": ["true"],
                "r-under-content-size": ["1000000"],
                "r-over-content-size": ["1000"],
                "r-action": ["extracted"],
                "d-self": ["true"],
                "r-streams": ["*,executable/windows/dos", "blob,archive/gzip,archive/bzip2"],
            },
            self.get_last_req_params(),
        )

    def test_post_and_get_new_content(self):
        # Tests the method where each item is posted individually and the server response with a ContentEntry
        p = sup.DummyPlugin(config={"events_url": self.server + "/generic", "data_url": self.server})
        sample_hashed = {
            "6a0e91ed5a21963490c9677934e6b99ae901f8eda7de3c1d9d52fa23d40bafb2": (["content"], b"fake data 1"),
            "f7d60cd635b71dc219222122a7bb5b14da90eb6b67c5bcd18234b7309582d2c2": (["content"], b"fake data 2"),
        }
        net = network.Network(p)
        net._post_data(
            "source",
            sample_hashed,
        )
        self.assertEqual(
            net._cached_file_data,
            {
                (
                    "source",
                    "content",
                    "6a0e91ed5a21963490c9677934e6b99ae901f8eda7de3c1d9d52fa23d40bafb2",
                ): azm.Datastream(
                    identify_version=1,
                    label=azm.DataLabel.CONTENT,
                    size=11,
                    magic="#TESTONLY",
                    mime="#TESTONLY",
                    file_format="#TEST/ONLY",
                    file_extension="tonly",
                    md5="69ea00f9c3104cced2e97c4662ba8b5c",
                    sha1="ac3ee7a4206a3b98702d20c925ecb6242ab64383",
                    sha256="6a0e91ed5a21963490c9677934e6b99ae901f8eda7de3c1d9d52fa23d40bafb2",
                    sha512="3f7e39f41ae6204a51b7a0103a1cccc9afb36be98f3afe2faef3ab3da0f0bb29"
                    "7d6a635c1cc06641d1ae40cee12f71671c85c492ee5b07d28ce2e1ed1bb7f193",
                ),
                (
                    "source",
                    "content",
                    "f7d60cd635b71dc219222122a7bb5b14da90eb6b67c5bcd18234b7309582d2c2",
                ): azm.Datastream(
                    identify_version=1,
                    label=azm.DataLabel.CONTENT,
                    size=11,
                    magic="#TESTONLY",
                    mime="#TESTONLY",
                    file_format="#TEST/ONLY",
                    file_extension="tonly",
                    md5="4af0be381e877439bc3b676519dea88f",
                    sha1="2ee5a439436eb9d5ee8c77eab994c854acbb7fab",
                    sha256="f7d60cd635b71dc219222122a7bb5b14da90eb6b67c5bcd18234b7309582d2c2",
                    sha512="85f2dc9828f5c359d22fa307405a5744f454c88a91f87b1d73ce97164869d9b5"
                    "18712abc36df09bba633a4f9143cb36c58b28a692c305a98a46ca38b23786a30",
                ),
            },
        )
        # Tests the newly posted content can be retrieved via a GET request.
        r = httpx.get(
            "%s/api/v3/stream/source/content/6a0e91ed5a21963490c9677934e6b99ae901f8eda7de3c1d9d52fa23d40bafb2"
            % self.server
        )
        r.raise_for_status()
        self.assertEqual(r.content, b"fake data 1")
        r = httpx.get(
            "%s/api/v3/stream/source/content/f7d60cd635b71dc219222122a7bb5b14da90eb6b67c5bcd18234b7309582d2c2"
            % self.server
        )
        r.raise_for_status()
        self.assertEqual(r.content, b"fake data 2")

    class _DPTestAckJob(sup.DummyPlugin):
        FEATURES = [
            Feature(name="feat1", desc="", type=azm.FeatureType.String),
            Feature(name="per_stream_feat", desc="", type=azm.FeatureType.String),
        ]
        SECURITY = "test_security_level"  # should be passed in result

    def test_ack_job(self):
        loop = monitor.Monitor(self._DPTestAckJob, {"events_url": self.server + "/test_data", "data_url": self.server})
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        _ = loop.run_once(local.gen_event(entity))
        dt = datetime.datetime.now(datetime.timezone.utc)
        event = loop._network.fetch_job()
        text_stream = b"Text data stream"
        text_id = DATA_HASH(text_stream).hexdigest()
        loop._network.ack_job(
            src=event,
            result=JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="1",
                        features={"feat1": {FeatureValue("This is a string value", label="label1")}},
                        info={
                            "list_field": [1, 2, 3],
                            "dict_field": {"foo": "bar", "int": 5},
                        },
                        data=[{"label": "text", "hash": text_id, "language": "english"}],
                    ),
                    Event(
                        sha256="1",
                        parent_sha256="b4b389c849d799d9331d5937cde7f0dfd297d76083242366cbef53b498cd6051",
                        features={"per_stream_feat": {FeatureValue('Set on "small content" stream')}},
                    ),
                ],
                data={
                    text_id: text_stream,
                },
                runtime=3,
                date_start=dt,
                date_end=dt + datetime.timedelta(seconds=3),
            ),
        )
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        print(out_event)
        # Override timestamps to make comparison easier
        out_event["timestamp"] = dt.isoformat()
        self.assertEqual(len(out_event["entity"]["results"]), 2, "There should be two events generated.")
        out_event["entity"]["results"][0]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][0]["source"]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][0]["source"]["path"][-1]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["source"]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["source"]["path"][-1]["timestamp"] = dt.isoformat()

        prepath = event.source.path
        event2 = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="runner-placeholder",
            action=azm.BinaryAction.Enriched,
            timestamp=dt,
            source=azm.Source(
                name="source",
                timestamp=dt,
                path=prepath
                + [
                    azm.PathNode(
                        action=azm.BinaryAction.Enriched,
                        timestamp=dt,
                        author=azm.Author(
                            category="plugin",
                            name="_DPTestAckJob",
                            version="1.0",
                            security="test_security_level",
                        ),
                        sha256="1234",
                    )
                ],
            ),
            author=azm.Author(
                category="plugin",
                name="_DPTestAckJob",
                version="1.0",
                security="test_security_level",
            ),
            entity=azm.BinaryEvent.Entity(
                sha256="1234",
                features=[
                    azm.FeatureValue(name="per_stream_feat", type="string", value='Set on "small content" stream'),
                ],
            ),
        )

        event1 = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="runner-placeholder",
            action=azm.BinaryAction.Augmented,
            timestamp=dt,
            source=azm.Source(
                name="source",
                timestamp=dt,
                path=prepath
                + [
                    azm.PathNode(
                        action=azm.BinaryAction.Augmented,
                        timestamp=dt,
                        author=azm.Author(
                            category="plugin",
                            name="_DPTestAckJob",
                            version="1.0",
                            security="test_security_level",
                        ),
                        sha256="1234",
                    )
                ],
            ),
            author=azm.Author(
                category="plugin",
                name="_DPTestAckJob",
                version="1.0",
                security="test_security_level",
            ),
            entity=azm.BinaryEvent.Entity(
                sha256="1234",
                datastreams=[
                    azm.Datastream(
                        identify_version=1,
                        file_format="#TEST/ONLY",
                        label=azm.DataLabel.CONTENT,
                        md5="5",
                        magic="mm",
                        mime="mt",
                        sha1="1",
                        sha256="b4b389c849d799d9331d5937cde7f0dfd297d76083242366cbef53b498cd6051",
                        sha512="512",
                        size=13,
                    ),
                    # alt-stream content won't be included
                    azm.Datastream(
                        identify_version=1,
                        label=azm.DataLabel.TEXT,
                        size=16,
                        language="english",
                        file_format="#TEST/ONLY",
                        file_extension="tonly",
                        mime="#TESTONLY",
                        magic="#TESTONLY",
                        md5="2fae0d09ab333884009fa812b6c4d097",
                        sha1="8805b0694e0128f2c7c22adb92d5515957161e87",
                        sha256="903e0722b8ebdfc5547af8ca1a9c8a76d0a7b067a7b1a561dfff13593e76d5a8",
                        sha512="718e7bc44d2e1396646fe014812c9af52d1b59c2858efbcc2d3bc096eda27289"
                        "2eac2b7bb2ac37bb9fc2f49b54b68f969f03d73e66e03a67c0a1a69ced48bce7",
                    ),
                ],
                features=[
                    azm.FeatureValue(name="feat1", type="string", value="This is a string value", label="label1"),
                ],
                info={
                    "list_field": [1, 2, 3],
                    "dict_field": {"foo": "bar", "int": 5},
                },
            ),
        )
        event.entity.features = []
        event.entity.datastreams = []
        event.entity.info = {}
        expected = dump(
            azm.StatusEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="runner-placeholder",
                timestamp=dt,
                author=azm.Author(
                    category="plugin",
                    name="_DPTestAckJob",
                    version="1.0",
                    security="test_security_level",
                ),
                entity=azm.StatusEvent.Entity(
                    input=event,
                    status=State.Label.COMPLETED,
                    results=[event1, event2],
                ),
            )
        )
        expected["entity"]["runtime"] = out_event["entity"]["runtime"]

        self.assertEqual(out_event["entity"]["results"][0], dump(event1))
        self.assertEqual(out_event["entity"]["results"][1], dump(event2))
        self.assertEqual(out_event, expected)

    class DPTestAckJobSecurityStringPassthrough(sup.DummyPlugin):
        FEATURES = [
            Feature(name="feat1", desc="", type=azm.FeatureType.String),
            Feature(name="per_stream_feat", desc="", type=azm.FeatureType.String),
        ]
        SECURITY = "test_security_level"  # should be passed in result

    def test_ack_job_security_string_passthrough(self):
        loop = monitor.Monitor(
            self.DPTestAckJobSecurityStringPassthrough,
            {"events_url": self.server + "/test_data", "data_url": self.server},
        )
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        _ = loop.run_once(local.gen_event(entity))[None]
        dt = datetime.datetime.now(datetime.timezone.utc)
        event = loop._network.fetch_job()
        event.author.security = "FISH"
        event.source.security = "FISH"
        # event.source.path[0].author.security = "FISH"
        text_stream = b"Text data stream"
        text_id = DATA_HASH(text_stream).hexdigest()
        loop._network.ack_job(
            src=event,
            result=JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="1",
                        parent_sha256="b4b389c849d799d9331d5937cde7f0dfd297d76083242366cbef53b498cd6051",
                        features={"per_stream_feat": {FeatureValue('Set on "small content" stream')}},
                    ),
                ],
                runtime=3,
                date_start=dt,
                date_end=dt + datetime.timedelta(seconds=3),
            ),
        )
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        # Override timestamps to make comparison easier
        out_event["timestamp"] = dt.isoformat()
        self.assertEqual(len(out_event["entity"]["results"]), 1)
        out_event["entity"]["results"][0]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][0]["source"]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][0]["source"]["path"][-1]["timestamp"] = dt.isoformat()

        event.entity.features = []
        event.entity.datastreams = []
        event.entity.info = {}
        prepath = event.source.path
        event1 = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="runner-placeholder",
            action=azm.BinaryAction.Enriched,
            timestamp=dt,
            source=azm.Source(
                name="source",
                timestamp=dt,
                path=prepath
                + [
                    azm.PathNode(
                        action=azm.BinaryAction.Enriched,
                        timestamp=dt,
                        author=azm.Author(
                            category="plugin",
                            name="DPTestAckJobSecurityStringPassthrough",
                            version="1.0",
                            security="test_security_level",
                        ),
                        sha256="1234",
                    )
                ],
                security="FISH",
            ),
            author=azm.Author(
                category="plugin",
                name="DPTestAckJobSecurityStringPassthrough",
                version="1.0",
                security="test_security_level",
            ),
            entity=azm.BinaryEvent.Entity(
                sha256="1234",
                features=[
                    azm.FeatureValue(name="per_stream_feat", type="string", value='Set on "small content" stream'),
                ],
            ),
        )
        expected = dump(
            azm.StatusEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="runner-placeholder",
                timestamp=dt,
                author=azm.Author(
                    category="plugin",
                    name="DPTestAckJobSecurityStringPassthrough",
                    version="1.0",
                    security="test_security_level",
                ),
                entity=azm.StatusEvent.Entity(
                    input=event,
                    status=State.Label.COMPLETED,
                    results=[event1],
                ),
            )
        )
        expected["entity"]["runtime"] = out_event["entity"]["runtime"]

        self.assertEqual(out_event["entity"]["results"][0], dump(event1))
        self.assertEqual(out_event, expected)

    class DPAckJobChildrenAndSecurityDict(sup.DummyPlugin):
        FEATURES = [
            Feature(name="sample_feature", desc="", type=azm.FeatureType.String),
        ]
        SECURITY = "ex_value TLP:WHITE"  # should be passed in result

    def test_ack_job_children_and_security_dict(self):
        """Test the API post output of an event with children and grandchildren, using a dict value for SECURITY"""

        loop = monitor.Monitor(
            self.DPAckJobChildrenAndSecurityDict, {"events_url": self.server + "/test_data", "data_url": self.server}
        )
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        _ = loop.run_once(local.gen_event(entity))[None]
        dt = datetime.datetime.now(datetime.timezone.utc)
        event = loop._network.fetch_job()
        child_data = b"foo"
        child_id = DATA_HASH(child_data).hexdigest()
        gc_data = b"bar"
        gc_id = DATA_HASH(gc_data).hexdigest()
        # Should not need to test 'per-stream' result output in combination with child/grandchildren,
        #  as the per-stream ack is tested above and the ack loop runs the same code over both (per-stream/normal)
        loop._network.ack_job(
            src=event,
            result=JobResult(
                **{
                    "state": State(State.Label.COMPLETED),
                    "events": [
                        {
                            "parent": {"sha256": ""},
                            "sha256": child_id,
                            "relationship": {"r": "v"},
                            "data": [{"label": "content", "hash": child_id}],
                        },
                        {
                            "parent": {
                                "parent": {"sha256": ""},
                                "sha256": child_id,
                                "relationship": {"r": "v"},
                            },
                            "sha256": gc_id,
                            "features": {"sample_feature": {FeatureValue("completed ok")}},
                            "relationship": {"r": "v2"},
                            "data": [{"label": "content", "language": "foo", "hash": gc_id}],
                        },
                    ],
                    "data": {child_id: child_data, gc_id: gc_data},
                    "runtime": 21,
                    "date_start": dt,
                    "date_end": dt + datetime.timedelta(seconds=21),
                }
            ),
        )
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        # Override timestamps to make comparison easier
        out_event["timestamp"] = dt.isoformat()
        self.assertEqual(len(out_event["entity"]["results"]), 2, "There should be two events generated.")
        out_event["entity"]["results"][0]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][0]["source"]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][0]["source"]["path"][-1]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["source"]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["source"]["path"][-1]["timestamp"] = dt.isoformat()
        out_event["entity"]["results"][1]["source"]["path"][-2]["timestamp"] = dt.isoformat()

        event.entity.features = []
        event.entity.datastreams = []
        event.entity.info = {}
        prepath = event.source.path
        expected = dump(
            azm.StatusEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="runner-placeholder",
                timestamp=dt,
                author=azm.Author(
                    category="plugin",
                    name="DPAckJobChildrenAndSecurityDict",
                    version="1.0",
                    security="ex_value TLP:WHITE",
                ),
                entity=azm.StatusEvent.Entity(
                    input=event,
                    status=State.Label.COMPLETED,
                    results=[
                        azm.BinaryEvent(
                            model_version=azm.CURRENT_MODEL_VERSION,
                            kafka_key="runner-placeholder",
                            action=azm.BinaryAction.Extracted,
                            timestamp=dt,
                            source=azm.Source(
                                name="source",
                                timestamp=dt,
                                path=prepath
                                + [
                                    azm.PathNode(
                                        action=azm.BinaryAction.Extracted,
                                        relationship={"r": "v"},
                                        timestamp=dt,
                                        author=azm.Author(
                                            category="plugin",
                                            name="DPAckJobChildrenAndSecurityDict",
                                            version="1.0",
                                            security="ex_value TLP:WHITE",
                                        ),
                                        sha256=child_id,
                                        size=3,
                                        file_format="#TEST/ONLY",
                                    )
                                ],
                            ),
                            author=azm.Author(
                                category="plugin",
                                name="DPAckJobChildrenAndSecurityDict",
                                version="1.0",
                                security="ex_value TLP:WHITE",
                            ),
                            entity=azm.BinaryEvent.Entity(
                                sha256=child_id,
                                datastreams=[
                                    azm.Datastream(
                                        identify_version=1,
                                        file_format="#TEST/ONLY",
                                        file_extension="tonly",
                                        label=azm.DataLabel.CONTENT,
                                        md5="acbd18db4cc2f85cedef654fccc4a4d8",
                                        magic="#TESTONLY",
                                        mime="#TESTONLY",
                                        sha1="0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33",
                                        sha256="2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
                                        sha512="f7fbba6e0636f890e56fbbf3283e524c6fa3204ae298382d624741d0dc663832"
                                        "6e282c41be5e4254d8820772c5518a2c5a8c0c7f7eda19594a7eb539453e1ed7",
                                        size=3,
                                    ),
                                ],
                                features=[
                                    azm.FeatureValue(name="file_format", type="string", value="#TEST/ONLY"),
                                    azm.FeatureValue(name="file_extension", type="string", value="tonly"),
                                    azm.FeatureValue(name="magic", type="string", value="#TESTONLY"),
                                    azm.FeatureValue(name="mime", type="string", value="#TESTONLY"),
                                ],
                                file_format="#TEST/ONLY",
                                file_extension="tonly",
                                mime="#TESTONLY",
                                magic="#TESTONLY",
                                md5="acbd18db4cc2f85cedef654fccc4a4d8",
                                sha1="0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33",
                                sha512="f7fbba6e0636f890e56fbbf3283e524c6fa3204ae298382d624741d0dc663832"
                                "6e282c41be5e4254d8820772c5518a2c5a8c0c7f7eda19594a7eb539453e1ed7",
                                size=3,
                            ),
                        ),
                        azm.BinaryEvent(
                            model_version=azm.CURRENT_MODEL_VERSION,
                            kafka_key="runner-placeholder",
                            action=azm.BinaryAction.Extracted,
                            timestamp=dt,
                            source=azm.Source(
                                name="source",
                                timestamp=dt,
                                path=prepath
                                + [
                                    azm.PathNode(
                                        action=azm.BinaryAction.Extracted,
                                        relationship={"r": "v"},
                                        timestamp=dt,
                                        author=azm.Author(
                                            category="plugin",
                                            name="DPAckJobChildrenAndSecurityDict",
                                            version="1.0",
                                            security="ex_value TLP:WHITE",
                                        ),
                                        sha256=child_id,
                                        size=3,
                                        file_format="#TEST/ONLY",
                                    ),
                                    azm.PathNode(
                                        action=azm.BinaryAction.Extracted,
                                        relationship={"r": "v2"},
                                        timestamp=dt,
                                        author=azm.Author(
                                            category="plugin",
                                            name="DPAckJobChildrenAndSecurityDict",
                                            version="1.0",
                                            security="ex_value TLP:WHITE",
                                        ),
                                        sha256=gc_id,
                                        size=3,
                                        file_format="#TEST/ONLY",
                                        language="foo",
                                    ),
                                ],
                            ),
                            author=azm.Author(
                                category="plugin",
                                name="DPAckJobChildrenAndSecurityDict",
                                version="1.0",
                                security="ex_value TLP:WHITE",
                            ),
                            entity=azm.BinaryEvent.Entity(
                                sha256=gc_id,
                                datastreams=[
                                    azm.Datastream(
                                        identify_version=1,
                                        file_format="#TEST/ONLY",
                                        file_extension="tonly",
                                        label=azm.DataLabel.CONTENT,
                                        language="foo",
                                        magic="#TESTONLY",
                                        mime="#TESTONLY",
                                        md5="37b51d194a7513e45b56f6524f2d51f2",
                                        sha1="62cdb7020ff920e5aa642c3d4066950dd1f01f4d",
                                        sha256="fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9",
                                        sha512="d82c4eb5261cb9c8aa9855edd67d1bd10482f41529858d925094d173fa662aa9"
                                        "1ff39bc5b188615273484021dfb16fd8284cf684ccf0fc795be3aa2fc1e6c181",
                                        size=3,
                                    ),
                                ],
                                features=[
                                    azm.FeatureValue(name="file_format", type="string", value="#TEST/ONLY"),
                                    azm.FeatureValue(name="file_extension", type="string", value="tonly"),
                                    azm.FeatureValue(name="magic", type="string", value="#TESTONLY"),
                                    azm.FeatureValue(name="mime", type="string", value="#TESTONLY"),
                                    azm.FeatureValue(name="sample_feature", type="string", value="completed ok"),
                                ],
                                file_format="#TEST/ONLY",
                                file_extension="tonly",
                                md5="37b51d194a7513e45b56f6524f2d51f2",
                                sha1="62cdb7020ff920e5aa642c3d4066950dd1f01f4d",
                                sha512="d82c4eb5261cb9c8aa9855edd67d1bd10482f41529858d925094d173fa662aa9"
                                "1ff39bc5b188615273484021dfb16fd8284cf684ccf0fc795be3aa2fc1e6c181",
                                size=3,
                                mime="#TESTONLY",
                                magic="#TESTONLY",
                            ),
                        ),
                    ],
                ),
            )
        )
        expected["entity"]["runtime"] = out_event["entity"]["runtime"]
        self.assertEqual(out_event, expected)

    class DPAckJobNewStreamOnly(sup.DummyPlugin):
        SECURITY = "test_security_level"  # should be passed in result

    def test_ack_job_new_stream_only(self):
        """Ensure if plugin only produces a new stream, it is still treated as an output result."""

        loop = monitor.Monitor(
            self.DPAckJobNewStreamOnly, {"events_url": self.server + "/test_data", "data_url": self.server}
        )
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        _ = loop.run_once(local.gen_event(entity))[None]
        dt = datetime.datetime.now(datetime.timezone.utc)
        event = loop._network.fetch_job()
        text_stream = b"Text data stream"
        text_id = DATA_HASH(text_stream).hexdigest()
        loop._network.ack_job(
            src=event,
            result=JobResult(
                **{
                    "state": State(State.Label.COMPLETED),
                    "data": {
                        text_id: text_stream,
                    },
                    "events": [
                        dict(
                            sha256="5",
                            data=[
                                dict(
                                    label=azm.DataLabel.TEXT,
                                    language="english",
                                    hash=text_id,
                                )
                            ],
                        )
                    ],
                    "runtime": 3,
                    "date_start": dt,
                    "date_end": dt + datetime.timedelta(seconds=3),
                }
            ),
        )
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        self.assertTrue(out_event["entity"].get("results"))

    class DPAckJobStatusId(sup.DummyPlugin):
        SECURITY = "test_security_level"  # should be passed in result

    def test_ack_job_status_id(self):
        loop = monitor.Monitor(
            self.DPAckJobStatusId, {"events_url": self.server + "/test_path", "data_url": self.server}
        )
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={})
        _ = loop.run_once(local.gen_event(entity))[None]
        dt = datetime.datetime.now(datetime.timezone.utc)
        event = loop._network.fetch_job()
        loop._network.ack_job(
            src=event,
            result=JobResult(
                **{
                    "state": State(State.Label.COMPLETED),
                    "data": {},
                    "events": [],
                    "runtime": 3,
                    "date_start": dt,
                    "date_end": dt + datetime.timedelta(seconds=3),
                }
            ),
        )
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        # Ensure that this hash matches what is tested against in the dispatcher
        self.assertEqual("generated-dummy-dequeued-id", out_event["entity"]["input"]["dequeued"])

    class DPRunloopPosting(sup.DummyPlugin):
        def execute(self, job):
            time.sleep(0.1)

    def test_runloop_posting(self):
        """Directly test monitoring process managing loop"""
        logging_queue = multiprocessing.Queue()

        p = Process(
            target=monitor._start_loop_coordinator,
            kwargs=dict(
                plugin=self.DPRunloopPosting,
                config=settings.parse_config(
                    self.DPRunloopPosting, {"events_url": self.server + "/null_sleep", "data_url": self.server}
                ),
                job_limit=None,
                log_level=logging.INFO,
                queue=self.dummy_queue,
                logging_queue=logging_queue,
            ),
        )
        p.start()
        while not p.is_alive():
            time.sleep(0.2)
        # Wait to make sure plugin enters main loop and executes a couple of times
        time.sleep(1.0)
        # Request a graceful shutdown
        p.terminate()
        self._wait_for_is_alive_to_be_false(p)
        # Expect that the plugin has terminated.
        self.assertFalse(p.is_alive())
        # Check that the last post was an ack job
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        j = r.json()[0]
        expected = dump(
            azm.StatusEvent(
                model_version=azm.CURRENT_MODEL_VERSION,
                kafka_key="runner-placeholder",
                timestamp=datetime.datetime.now(tz=datetime.timezone.utc),  # Will be replaced with correct value below
                author=azm.Author(
                    category="plugin",
                    name="DPRunloopPosting",
                    version="1.0",
                ),
                entity=azm.StatusEvent.Entity(
                    input=azm.BinaryEvent(
                        model_version=azm.CURRENT_MODEL_VERSION,
                        kafka_key="test-dummy",
                        dequeued="test-dummy-dequeued",
                        action=azm.BinaryAction.Enriched,
                        timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                        # hash="dbb2466f01612c0e632e3ca8bfc1de02",
                        source=azm.Source(
                            name="source",
                            path=[],
                            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                        ),
                        author=azm.Author(name="TestServer", category="blah"),
                        entity=azm.BinaryEvent.Entity(sha256="1234", datastreams=[], features=[]),
                    ),
                    status=State.Label.COMPLETED_EMPTY,
                    results=[],  # null_sleep does not produce features or children
                ),
            )
        )
        print(json.dumps(j))
        # Copy the values that vary each run
        expected["timestamp"] = j["timestamp"]
        expected["entity"]["runtime"] = j.get("entity", {}).get("runtime")
        self.assertEqual(j, expected)

    class DPRunloopWithHeartbeat(sup.DummyPlugin):
        def execute(self, job: Job):
            time.sleep(2.5)
            self.add_feature_values("example_string", "test")

    def test_runloop_with_heartbeat(self):
        # Clean any requests
        r = httpx.get("%s/mock/get_var/all_requests" % self.server)

        # Send heartbeat ever 1 sec
        loop = monitor.Monitor(
            self.DPRunloopWithHeartbeat,
            {"events_url": self.server + "/test_data", "heartbeat_interval": 1, "data_url": self.server},
        )
        test_data = b"This is test data that should be fetched by the runner"
        loop._network._post_data(
            "source",
            {DATA_HASH(test_data).hexdigest(): (["content"], test_data)},
        )
        loop.run_loop(1)

        # Check that heartbeats have been sent
        r = httpx.get("%s/mock/get_var/all_requests" % self.server)

        r.raise_for_status()
        out_evts: list[azm.StatusEvent] = [
            azm.StatusEvent(**ev[0]) for ev in r.json() if ev[0].get("entity", {}).get("input")
        ]
        print("---------START-------------")
        print(out_evts)
        print("---------START-------------")
        # Allow for either 1 or 2 heartbeats
        self.assertLessEqual(2, len(out_evts))
        self.assertGreaterEqual(3, len(out_evts))

        for out_evt in out_evts:
            self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        for out_evt in out_evts[:-1]:
            self.assertEqual(State.Label.HEARTBEAT, out_evt.entity.status)

        self.assertEqual(out_evts[-1].entity.status, State.Label.COMPLETED)
        self.assertEqual(
            out_evts[-1].entity.results[0].entity.features,
            [azm.FeatureValue(name="example_string", type="string", value="test")],
        )

    class DPRunloopWithContent(sup.DummyPlugin):
        def execute(self, job: Job):
            if job.get_all_data():
                pass
            else:
                # Alternative way of failing test to avoid serialising self
                raise coordinator.CriticalError("get_all_data is not truthy!")
            lengths = []
            for ds in job.get_all_data():
                if not isinstance(ds, StorageProxyFile):
                    # Alternative way of failing test to avoid serialising self
                    raise coordinator.CriticalError("Provided data is not a StorageProxyFile!")
                lengths.append(len(ds.read()))
            self.add_feature_values("example_int", lengths)

    def test_runloop_with_content(self):
        loop = monitor.Monitor(
            self.DPRunloopWithContent, {"events_url": self.server + "/test_data", "data_url": self.server}
        )
        test_data = b"This is test data that should be fetched by the runner"
        loop._network._post_data(
            "source",
            {DATA_HASH(test_data).hexdigest(): (["content"], test_data)},
        )
        loop.run_loop(1)
        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.COMPLETED,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )
        self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
        self.assertEqual(
            out_evt.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_int", type="integer", value="13"),
                azm.FeatureValue(name="example_int", type="integer", value="54"),
            ],
        )

    class DPRunloopMultipleIterations(sup.DummyPlugin):
        def execute(self, job: Job):
            if job.get_all_data():
                pass
            else:
                # Alternative way of failing test to avoid serialising self
                raise coordinator.CriticalError("get_all_data is not truthy!")
            lengths = []
            for ds in job.get_all_data():
                if not isinstance(ds, StorageProxyFile):
                    # Alternative way of failing test to avoid serialising self
                    raise coordinator.CriticalError("Provided data is not a StorageProxyFile!")
                lengths.append(len(ds.read()))
            self.add_feature_values("example_int", lengths)

    def test_runloop_multiple_iterations(self):
        """Ensure the plugin can run in a loop and post more than one response."""
        loop = monitor.Monitor(
            self.DPRunloopMultipleIterations, {"events_url": self.server + "/test_data", "data_url": self.server}
        )
        loop.time_to_wait_between_checks = 0.5
        test_data = b"This is test data that should be fetched by the runner"
        loop._network._post_data(
            "source",
            {DATA_HASH(test_data).hexdigest(): (["content"], test_data)},
        )
        request_count = httpx.get("%s/mock/get_var/fetch_count" % self.server)
        inital_request_count = int(request_count.content.decode())
        loop.run_loop(5)
        request_count = httpx.get("%s/mock/get_var/fetch_count" % self.server)
        requests_in_interval = int(request_count.content.decode()) - inital_request_count
        # Ensure there was at least five requests.
        self.assertGreaterEqual(requests_in_interval, 5)

        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.COMPLETED,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )
        self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
        self.assertEqual(
            out_evt.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_int", type="integer", value="13"),
                azm.FeatureValue(name="example_int", type="integer", value="54"),
            ],
        )

    def test_runloop_depth_limit_reached(self):
        # No specific authors to filter but depth has reached config
        loop = monitor.Monitor(
            sup.DummyPlugin, {"events_url": self.server + "/depth_2", "plugin_depth_limit": 2, "data_url": self.server}
        )
        loop.run_loop(1)
        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.OPT_OUT,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )

        # No specific authors to filter but depth has reached config
        loop = monitor.Monitor(
            sup.DummyPlugin,
            {
                "server": self.server + "/depth_dupe",
                "plugin_depth_limit": 2,
            },
        )
        loop.run_loop(1)
        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.OPT_OUT,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )

    class DPRunloopWithWatch(sup.DummyPlugin):
        def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
            super().__init__(config)
            filepath = self.cfg.test_input_data.get("filepath")
            with open(os.path.join(filepath, "tmp.txt"), "r") as f:
                self.retval = f.read()

        def execute(self, job):
            self.add_feature_values("example_string", self.retval)

    def test_runloop_with_watch(self):
        with tempfile.TemporaryDirectory() as filepath:
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("1")

            loop = monitor.Monitor(
                self.DPRunloopWithWatch,
                {
                    "events_url": self.server + "/test_data",
                    "watch_path": filepath,
                    "watch_wait": 0,
                    "data_url": self.server,
                    "test_input_data": {"filepath": filepath},
                },
            )
            # test_data = b"This is test data that should be fetched by the runner"
            # loop._network._post_data({DATA_HASH(test_data).hexdigest(): test_data})
            loop.run_loop(1)
            # Check that the ack contains the correct features
            r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
            r.raise_for_status()
            out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
            self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
            self.assertEqual(
                out_evt.entity.status,
                State.Label.COMPLETED,
                "Unexpected result: %s:%s (%s)"
                % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
            )
            self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
            self.assertEqual(
                out_evt.entity.results[0].entity.features,
                [
                    azm.FeatureValue(name="example_string", type="string", value="1"),
                ],
            )

            # second run with new values
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("2")

            # idk how long it takes to pick up changes
            time.sleep(1)

            loop.run_loop(1)
            # Check that the ack contains the correct features
            r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
            r.raise_for_status()
            out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
            self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
            self.assertEqual(
                out_evt.entity.status,
                State.Label.COMPLETED,
                "Unexpected result: %s:%s (%s)"
                % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
            )
            self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
            # must have updated value here
            self.assertEqual(
                out_evt.entity.results[0].entity.features,
                [
                    azm.FeatureValue(name="example_string", type="string", value="2"),
                ],
            )
        # kill the watchdog threads to reduce chance of weird stuff during testing
        del loop

    def test_runloop_with_no_watch(self):
        with tempfile.TemporaryDirectory() as filepath:
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("1")

            class DPRunlookWithNoWatch(sup.DummyPlugin):
                def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
                    super().__init__(config)
                    with open(os.path.join(filepath, "tmp.txt"), "r") as f:
                        self.retval = f.read()

                def execute(self, job):
                    self.add_feature_values("example_string", self.retval)

            # Have to run with coordinator, because monitor would restart the coordinator between iterations.
            loop = coordinator.Coordinator(
                DPRunlookWithNoWatch,
                settings.parse_config(
                    DPRunlookWithNoWatch, {"events_url": self.server + "/test_data", "data_url": self.server}
                ),
            )
            # test_data = b"This is test data that should be fetched by the runner"
            # loop._network._post_data({DATA_HASH(test_data).hexdigest(): test_data})
            with self.assertRaises(SystemExit) as e:
                loop.run_loop(queue=self.dummy_queue, job_limit=1)

            # Verify clean exit
            self.assertEqual(e.exception.code, 0)

            # Check that the ack contains the correct features
            r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
            r.raise_for_status()
            out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
            self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
            self.assertEqual(
                out_evt.entity.status,
                State.Label.COMPLETED,
                "Unexpected result: %s:%s (%s)"
                % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
            )
            self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
            self.assertEqual(
                out_evt.entity.results[0].entity.features,
                [
                    azm.FeatureValue(name="example_string", type="string", value="1"),
                ],
            )

            # second run with new values
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("2")

            # idk how long it takes to pick up changes
            time.sleep(1)
            with self.assertRaises(SystemExit) as e:
                loop.run_loop(queue=self.dummy_queue, job_limit=1)

            # Verify clean exit
            self.assertEqual(e.exception.code, 0)

            # Check that the ack contains the correct features
            r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
            r.raise_for_status()
            out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
            self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
            self.assertEqual(
                out_evt.entity.status,
                State.Label.COMPLETED,
                "Unexpected result: %s:%s (%s)"
                % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
            )
            self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
            print("_______________________________________A_______________________________________")
            print(out_evt.entity.results[0].entity.features)
            print("_______________________________________B_______________________________________")
            # must not have updated value here
            self.assertEqual(
                out_evt.entity.results[0].entity.features,
                [
                    azm.FeatureValue(name="example_string", type="string", value="1"),
                ],
            )

    class DPRunloopUTF8(sup.DummyPlugin):
        def execute(self, job: Job):
            lengths = []
            if job.get_all_data():
                pass
            else:
                # A way of crashing the test without assertions
                raise coordinator.CriticalError("get_all_data is not truthy!")
            for ds in job.get_all_data():
                if not isinstance(ds, StorageProxyFile):
                    raise coordinator.CriticalError("Provided data is not a StorageProxyFile!")
                lengths.append(len(ds.read()))
            self.add_feature_values("example_string", self.cfg.test_input_data.get("example_string"))

    def test_runloop_utf8(self):
        # ensure that utf8 is encoded and decoded correctly
        # different length when encoded
        text = "中英文均可"
        self.assertEqual(len(text), 5)
        self.assertEqual(len(text.encode()), 15)

        loop = monitor.Monitor(
            TestBasePluginLive.DPRunloopUTF8,
            {
                "events_url": self.server + "/test_data",
                "data_url": self.server,
                "test_input_data": {"example_string": text},
            },
        )
        test_data = b"This is test data that should be fetched by the runner"
        loop._network._post_data(
            "source",
            {DATA_HASH(test_data).hexdigest(): (["content"], test_data)},
        )
        loop.run_loop(1)
        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.COMPLETED,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )
        self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
        self.assertEqual(
            out_evt.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_string", type="string", value=text),
            ],
        )

    class DPRunloopSurrogateCharacters(sup.DummyPlugin):
        def execute(self, job: Job):
            lengths = []
            if job.get_all_data():
                pass
            else:
                # A way of crashing the test without assertions
                raise coordinator.CriticalError("get_all_data is not truthy!")
            for ds in job.get_all_data():
                if not isinstance(ds, StorageProxyFile):
                    raise coordinator.CriticalError("Provided data is not a StorageProxyFile!")
                lengths.append(len(ds.read()))
            self.add_feature_values("example_string", self.cfg.test_input_data.get("text"))

    def test_runloop_surrogate_characters(self):
        # ensure that utf8 is encoded and decoded correctly
        # different length when encoded
        text = "\ud83d\ude4f"
        expected_fv_text = "\\ud83d\\ude4f"
        self.assertEqual(len(text), 2)
        self.assertEqual(len(text.encode(errors="backslashreplace")), 12)

        loop = monitor.Monitor(
            self.DPRunloopSurrogateCharacters,
            {"events_url": self.server + "/test_data", "data_url": self.server, "test_input_data": {"text": text}},
        )
        test_data = b"This is test data that should be fetched by the runner"
        loop._network._post_data(
            "source",
            {DATA_HASH(test_data).hexdigest(): (["content"], test_data)},
        )
        loop.run_loop(1)
        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.COMPLETED,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )
        self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
        self.assertEqual(
            out_evt.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_string", type="string", value=expected_fv_text),
            ],
        )

    def test_too_many_events(self):
        loop = monitor.Monitor(sup.DummyPlugin, {"events_url": self.server + "/too_many", "data_url": self.server})
        self.assertRaisesRegex(ValueError, "2 events fetched by dispatcher, only 1 allowed", loop._network.fetch_job)

    def test_use_spawn_mp_plugin(self):
        """Test using spawn multiprocessing setting works with yarax."""
        in_event = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="test-dummy",
            dequeued="test-dummy-dequeued",
            action=azm.BinaryAction.Sourced,
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            source=azm.Source(
                name="source",
                path=[],
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            ),
            author=azm.Author(name="TestServer", category="blah"),
            entity=azm.BinaryEvent.Entity(sha256="1234", datastreams=[], features=[]),
        )

        loop = monitor.Monitor(YaraxDP, {})
        final_result = loop.run_once(in_event)
        print(final_result)
        self.assertEqual(
            final_result[None].events[0].features,
            {"example_string": [FV(value="test")]},
        )

    class DPRunMultipleInstancesConcurrently(sup.DummyPlugin):
        SETTINGS = add_settings(
            request_retry_count=0,
            concurrent_plugin_instances=5,
        )  # Don't retry failed requests when testing

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            self.received_jobs = 0

        def execute(self, job: Job):
            if self.received_jobs and self.received_jobs == 1:
                # Sleep before completing second job.
                time.sleep(1)
            elif self.received_jobs and self.received_jobs > 1:
                # Ran more than 2 jobs where only a max of 2 is allowed.
                raise coordinator.CriticalError("Instance received more than one job and this is unexpected!")

            self.received_jobs += 1
            if job.get_all_data():
                pass
            else:
                # Alternative way of failing test to avoid serialising self
                raise coordinator.CriticalError("get_all_data is not truthy!")
            lengths = []
            for ds in job.get_all_data():
                if not isinstance(ds, StorageProxyFile):
                    # Alternative way of failing test to avoid serialising self
                    raise coordinator.CriticalError("Provided data is not a StorageProxyFile!")
                lengths.append(len(ds.read()))
            self.add_feature_values("example_int", lengths)

    def test_run_multiple_plugin_instances_concurrently(self):
        """Run 5 instances of a plugin concurrently and ensure they get at most 2 jobs.

        There should be 5 instances running a total of 5 jobs.
        The plugins should all run at least once and may run twice.
        But must not run more than twice. (or else there might just be one instance of the plugin.)
        The reason for this is, the plugin instances will all process their first job at roughly the same time.
        They may not ALL be done by the time coordinator checks if the job limit is reached.
        Once the job limit is reached all plugins will have one at least once.
        """
        loop = monitor.Monitor(
            self.DPRunMultipleInstancesConcurrently,
            {"events_url": self.server + "/test_data", "data_url": self.server},
        )
        loop.time_to_wait_between_checks = 0.5
        test_data = b"This is test data that should be fetched by the runner"
        loop._network._post_data(
            "source",
            {DATA_HASH(test_data).hexdigest(): (["content"], test_data)},
        )
        request_count = httpx.get("%s/mock/get_var/fetch_count" % self.server)
        inital_request_count = int(request_count.content.decode())
        loop.run_loop(5)
        request_count = httpx.get("%s/mock/get_var/fetch_count" % self.server)
        requests_in_interval = int(request_count.content.decode()) - inital_request_count
        # Ensure there was at least five requests.
        self.assertGreaterEqual(requests_in_interval, 5)

        # Check that the ack contains the correct features
        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[0])
        self.assertIsInstance(out_evt.entity, azm.StatusEvent.Entity)
        self.assertEqual(
            out_evt.entity.status,
            State.Label.COMPLETED,
            "Unexpected result: %s:%s (%s)" % (out_evt.entity.status, out_evt.entity.error, out_evt.entity.message),
        )
        self.assertEqual(len(out_evt.entity.results), 1, "Expected one result entity; got:\n%s" % out_evt)
        self.assertEqual(
            out_evt.entity.results[0].entity.features,
            [
                azm.FeatureValue(name="example_int", type="integer", value="13"),
                azm.FeatureValue(name="example_int", type="integer", value="54"),
            ],
        )


class YaraxDP(sup.DummyPlugin):
    """Create a plugin that uses yara-x to ensure it works in azul-runner.

    This is because yara-x fails to function when using certain multiprocessing options
    and always fails when using threading.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rules = yara_x.compile("""
        rule test { 
            strings: 
            $a = "fetched" 
            condition: 
            $a
        }""")

    def execute(self, job: Job):
        result = self.rules.scan(b"This is test data that should be fetched by the runner")
        for rule_hit in result.matching_rules:
            self.add_feature_values("example_string", rule_hit.identifier)



from __future__ import annotations

import contextlib
import datetime
import json
import logging
import multiprocessing
import os
import signal
import tempfile
import time
import unittest
from typing import ClassVar
from unittest.mock import patch

import httpx
import psutil
import pytest
from azul_bedrock import models_network as azm

from azul_runner import State, coordinator, monitor, settings
from azul_runner.settings import add_settings
from tests import plugin_support as sup

from . import mock_dispatcher as md
from .test_plugin_timeout import DummySleepPlugin, TestPluginTimeouts


def _proxy_run_loop_monitor(server: str, *args):
    """Raise a termination signal"""
    loop = monitor.Monitor(
        DummySleepPlugin,
        {"events_url": server + "/test_data", "data_url": server, "delay": 3},
    )
    loop.run_loop()


def _proxy_run_loop_coordinator(server: str, dummy_queue):
    """Raise a termination signal"""
    loop = coordinator.Coordinator(
        DummySleepPlugin,
        settings.Settings(events_url=server + "/test_data", data_url=server, delay=1, delay_after_exception=0),
    )
    loop.run_loop(queue=dummy_queue)


class CustomTestException(Exception):
    """Unique exception that is only raised in tests."""

    pass


def dump(x):
    return json.loads(x.model_dump_json(exclude_defaults=True))


class TestPluginTerminated(unittest.TestCase):
    """
    Tests a plugin stops when it recieves an appropriate signal.
    """

    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server, suitable for passing to a plugin's config['server']
    dummy_log_handler: sup.DummyLogHandler  # Set for each instance in setUp()
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()
        cls.editor = md.Editor(cls.server)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()

    def setUp(self):
        self.basic_input_event = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="test-dummy",
            dequeued="test-dummy-dequeued",
            action=azm.BinaryAction.Sourced,
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            source=azm.Source(
                name="source",
                path=[],
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            ),
            author=azm.Author(name="TestServer", category="blah"),
            entity=azm.BinaryEvent.Entity(sha256="1234", datastreams=[], features=[]),
        )
        self.dummy_queue: multiprocessing.Queue = multiprocessing.Queue()

    @pytest.mark.timeout(20)
    def test_sigterm_monitor(self):
        """Test to see if child processes are killed when a sigterm is sent to the parent process (monitor)"""
        process_ref = multiprocessing.Process(
            target=_proxy_run_loop_monitor,
            args=(self.server),
        )
        process_ref.start()

        # Wait until parent id is available.
        parent = None
        for i in range(20):
            time.sleep(0.5)
            with contextlib.suppress(psutil.NoSuchProcess):
                parent = psutil.Process(process_ref.pid)
            if i >= 19:
                raise Exception("cannot find parent process after waiting, something is wrong!")
            if parent:
                break

        all_pids = [parent.pid]
        for child_processes in parent.children(recursive=True):
            all_pids.append(child_processes.pid)

        # Should be at least 1 child process
        self.assertGreaterEqual(len(all_pids), 1)
        # Verify processes are running
        for p in all_pids:
            cur_process = psutil.Process(p)
            self.assertTrue(cur_process.is_running())

        # Send sigterm
        process_ref.terminate()
        # Wait up to 20 seconds for process to exit
        process_ref.join(20)

        self.assertFalse(process_ref.is_alive())

        # Verify all child processes are exited.
        # (this takes slightly longer than the parent process, due to propagation time, so continually check.)
        one_alive = True
        while one_alive:
            one_alive = False
            for p in all_pids:
                with contextlib.suppress(Exception):
                    cur_process = psutil.Process(p)
                    time.sleep(0.5)
                    one_alive = True

        for p in all_pids:
            with self.assertRaises(psutil.NoSuchProcess):
                cur_process = psutil.Process(p)

    @pytest.mark.timeout(20)
    def test_sigterm_coordinator(self):
        """Test to see if coordinator completes it's last job and then exits when receiving a SIGTERM.

        As opposed to just accepting the SIGTERM and exiting immediately.
        """
        p = multiprocessing.Process(
            target=_proxy_run_loop_coordinator,
            args=(self.server, self.dummy_queue),
        )
        p.start()
        time.sleep(2)
        # Send sigterm
        p.terminate()
        # Wait up to 20 seconds for process to exit
        p.join(20)
        # Verify that jobs were processed and that the last message to the dummy_queue is None.
        # None should be the last value because this is placed on the queue during a fetch to dispatcher.
        # It indicates that the plugin didn't immediately exit and completed successfully before accepting the SIGTERM.
        # This is because None is only added to the queue just before the fetch from dispatcher.
        # This test would fail intermittently if that was happening by chance
        queue_val = 1
        num_none_vals = 0
        num_jobs = 0
        while not self.dummy_queue.empty():
            queue_val = self.dummy_queue.get()
            if queue_val:
                num_jobs += 1
            else:
                num_none_vals += 1
            # The queue has a None event, Job, None event, Job each iteration.
            # None events indicate a job is done, where Job events indicate a job is being processed.

        self.assertIsNone(queue_val)
        self.assertGreaterEqual(2, num_jobs)
        self.assertGreaterEqual(3, num_none_vals)



"""Common logging utilities for azul-runner."""

import logging
import multiprocessing
import sys
from logging import handlers
from typing import Optional, TypeAlias

AZUL_RUNNER_LOGGER_NAME = "azul_runner"
LogLevel: TypeAlias = int | str


def setup_logger(log_level: LogLevel, queue: Optional[multiprocessing.Queue] = None):
    """Setup the loggers for azul runner."""
    # use log level settings
    rootlogger = logging.getLogger()
    # Set all loggers to error/warning
    rootlogger.setLevel(log_level)

    # Clear existing handlers
    for old_handle in rootlogger.handlers:
        rootlogger.removeHandler(old_handle)

    # Set the handler and formatter for azul_runner logggers.
    if queue:
        # Logging Queue is in use, send all logs to the parent process via the queue and don't log to stderr.
        h3 = handlers.QueueHandler(queue)
        h3.setLevel(log_level)
        rootlogger.addHandler(h3)
    else:
        # Root process is configured to log to stderr.
        h = logging.StreamHandler(sys.stderr)
        h.setLevel(log_level)
        h.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s %(levelname)-8s %(name)-30s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S%z"
            )
        )
        rootlogger.addHandler(h)

    # Set HTTPX logging to debug to avoid logging spam in production
    logging.getLogger("httpx").setLevel("DEBUG")
    logging.getLogger("httpcore").setLevel("DEBUG")


class AddLoggingQueueListener:
    """Handle adding and removing a logging queue handler."""

    def __init__(self, queue: multiprocessing.Queue, logger: logging.Logger):
        """Handle adding and removing a logging queue handler."""
        self.queue = queue
        self.logger = logger

    def __enter__(self):
        """Open the queue listener."""
        self.q_listen = handlers.QueueListener(self.queue, *self.logger.handlers)
        self.q_listen.start()

    def __exit__(self, *args, **kwargs):
        """Close the Queue listener to stop log capture."""
        self.q_listen.stop()



"""Local event generation."""

import datetime
import hashlib
import typing
import unittest

from azul_bedrock import identify
from azul_bedrock import models_network as azm

from . import storage
from .settings import Settings
from .storage import StorageProxyFile

unittest.util._MAX_LENGTH = 2000


def validate_streams(datastreams: list[StorageProxyFile], plugin_settings: Settings):
    """Verify that a set of data streams matches input reqs specified by plugin, or raise AssertionError."""
    # calculate what data the job provided in same format as INPUT_DATA
    available: dict[str, set[str]] = {}
    magics = set()
    mimes = set()

    for ds in datastreams:
        types = ds.file_info.file_format.split(";")[0]
        available.setdefault(ds.file_info.label, set()).add(types)
        available.setdefault("*", set()).add(types)
        magics.add(ds.file_info.magic)
        mimes.add(ds.file_info.mime)

    # check INPUT_DATA against job provided
    is_matching_type = False
    for content_label, plugin_allowed_formats in plugin_settings.filter_data_types.items():
        if content_label not in available:
            break
        # if no required type is available
        if plugin_allowed_formats:
            for allowed_type in available[content_label]:
                for allowed_format in plugin_allowed_formats:
                    if allowed_type.startswith(allowed_format):
                        is_matching_type = True
                        break
        # Type found
        if is_matching_type:
            return

    if any(len(list_of_values) > 0 for list_of_values in plugin_settings.filter_data_types.values()):
        raise AssertionError(
            f"Provided file format {available} does not meet plugin requirements {plugin_settings.filter_data_types}\n"
            f"magic: {sorted(magics)}\n"
            f"mime types: {sorted(mimes)}"
        )


def gen_api_content(binaryio: typing.BinaryIO, label: azm.DataLabel = azm.DataLabel.CONTENT):
    """Generate api content from bytes."""
    # get last offset and assume that is correct size
    binaryio.seek(0, 2)
    size = binaryio.tell()
    binaryio.seek(0)

    magic, mime, file_format, file_extension = identify.from_buffer(binaryio.read())
    binaryio.seek(0)
    return azm.Datastream(
        label=label,
        size=size,
        md5=storage.calc_stream_hash(binaryio, hashlib.md5),
        sha1=storage.calc_stream_hash(binaryio, hashlib.sha1),
        sha256=storage.calc_stream_hash(binaryio, hashlib.sha256),
        sha512=storage.calc_stream_hash(binaryio, hashlib.sha512),
        mime=mime,
        magic=magic,
        file_format=file_format,
        file_extension=file_extension,
    )


def gen_event(entity: azm.BinaryEvent.Entity):
    """Generate a simple network event."""
    return azm.BinaryEvent(
        model_version=azm.CURRENT_MODEL_VERSION,
        kafka_key="test",
        dequeued="test-response",
        action=azm.BinaryAction.Extracted,
        timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
        source=azm.Source(
            name="source",
            path=[
                azm.PathNode(
                    author=azm.Author(name="TestServer", category="plugin"),
                    action=azm.BinaryAction.Extracted,
                    timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                    sha256=entity.sha256,
                ),
            ],
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
        ),
        author=azm.Author(name="TestServer", category="plugin"),
        entity=entity,
    )



from __future__ import annotations

import datetime
import json
import os
import tempfile
import time
import unittest
from typing import ClassVar

import httpx
import pytest
from azul_bedrock import models_network as azm

from azul_runner import State, monitor
from azul_runner.settings import add_settings
from tests import plugin_support as sup

from . import mock_dispatcher as md
from .test_plugin_timeout import DummySleepPlugin, TestPluginTimeouts


class CustomTestException(Exception):
    """Unique exception that is only raised in tests."""

    pass


def dump(x):
    return json.loads(x.model_dump_json(exclude_defaults=True))


class SleepMultiPlugin(sup.DummyPlugin):
    SETTINGS = add_settings(memory_file_path=(str, ""))

    def mp_callback(self, job):
        "Write 100% to memory usage and then sleep."
        time.sleep(1)
        # Write to the file if it exists (if it doesn't the process has been terminated)
        if os.path.exists(self.cfg.memory_file_path):
            with open(self.cfg.memory_file_path, "w") as f:
                f.write("100")
                f.flush()
                f.seek(0)
        time.sleep(1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Write to the file if it exists (if it doesn't the process has been terminated)
        if os.path.exists(self.cfg.memory_file_path):
            with open(self.cfg.memory_file_path, "w") as f:
                f.write("0")
                f.flush()
                f.seek(0)

        self.register_multiplugin("MultiPluginName", "1.0", self.mp_callback)

    def execute(self, entity):
        pass


class TestPluginOom(unittest.TestCase):
    """
    Tests the Out of Memory functionality of Plugin.run_loop(), using the mock server.
    """

    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server, suitable for passing to a plugin's config['server']
    dummy_log_handler: sup.DummyLogHandler  # Set for each instance in setUp()
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()

    def setUp(self):
        self.cur_mem_summary_file_path = tempfile.NamedTemporaryFile("w+", prefix="dontdelete")
        # memory summary file
        self.cur_mem_summary_file_path.write(
            "\n".join(
                [
                    "anon 1228517376",
                    "file 19820544",
                    "kernel_stack 704512",
                    "pagetables 4546560",
                    "percpu 1152",
                    "sock 69632",
                    "shmem 262144",
                    "file_mapped 1372160",
                    "file_dirty 57344",
                    "file_writeback 0",
                    "swapcached 0",
                    "anon_thp 811597824",
                    "file_thp 0",
                    "shmem_thp 0",
                    "inactive_anon 1257385984",
                    "active_anon 49152",
                    "inactive_file 10",
                    "active_file 7118848",
                    "unevictable 0",
                    "slab_reclaimable 4810728",
                    "slab_unreclaimable 1600464",
                    "slab 6411192",
                    "workingset_refault_anon 0",
                    "workingset_refault_file 0",
                    "workingset_activate_anon 0",
                    "workingset_activate_file 0",
                    "workingset_restore_anon 0",
                    "workingset_restore_file 0",
                    "workingset_nodereclaim 0",
                    "pgfault 2343992",
                    "pgmajfault 22",
                    "pgrefill 0",
                    "pgscan 0",
                    "pgsteal 0",
                    "pgactivate 1786",
                    "pgdeactivate 0",
                    "pglazyfree 0",
                    "pglazyfreed 0",
                    "thp_fault_alloc 2343",
                    "thp_collapse_alloc 508",
                ]
            )
        )
        self.cur_mem_summary_file_path.flush()
        self.cur_mem_summary_file_path.seek(0)
        # View contents of file.
        # print(self.cur_mem_summary_file_path.read())
        # self.cur_mem_summary_file_path.seek(0)

        # Add a prefix to avoid deletion of temporary files
        self.max_mem_file = tempfile.NamedTemporaryFile("w+", prefix="dontdelete")
        self.max_mem_file.write("100")
        self.max_mem_file.flush()
        self.max_mem_file.seek(0)
        # Add a prefix to avoid deletion of temporary files
        self.cur_mem_file = tempfile.NamedTemporaryFile("w+", prefix="dontdelete")
        self.cur_mem_file.write("0")
        self.cur_mem_file.flush()
        self.cur_mem_file.seek(0)
        self.default_config = {
            # Mem limit settings
            "enable_mem_limits": True,
            "used_mem_warning_frac": 0.8,
            "used_mem_force_exit_frac": 0.9,
            "max_mem_file_path": self.max_mem_file.name,
            "cur_mem_file_path": self.cur_mem_file.name,
            "cur_mem_summary_file_path": self.cur_mem_summary_file_path.name,
            "mem_poll_frequency_milliseconds": 1000,
        }

        self.basic_input_event = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="test-dummy",
            dequeued="test-dummy-dequeued",
            action=azm.BinaryAction.Sourced,
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            source=azm.Source(
                name="source",
                path=[],
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            ),
            author=azm.Author(name="TestServer", category="blah"),
            entity=azm.BinaryEvent.Entity(sha256="1234", datastreams=[], features=[]),
        )

    def tearDown(self) -> None:
        try:
            self.max_mem_file.close()
        except Exception:
            pass
        try:
            self.cur_mem_summary_file_path.close()
        except Exception:
            pass
        try:
            self.cur_mem_file.close()
        except Exception:
            pass

        return super().tearDown()

    # ############### #
    # #### Tests #### #
    # ############### #
    @pytest.mark.timeout(10)
    def test_normal_plugin_run(self):
        """Check that the plugin runs and returns successfully with a memory limit set when it stays in that limit."""
        loop = monitor.Monitor(DummySleepPlugin, self.default_config.copy())
        result = loop.run_once(self.basic_input_event)

        print(result[None])
        self.assertEqual(result[None].state, State(State.Label.COMPLETED_EMPTY))

    @pytest.mark.timeout(5)
    def test_oom_failure(self):
        """Tests that the plugin exits with an exception if memory usage is over the error threshold."""
        self.cur_mem_file.write("100")
        self.cur_mem_file.flush()
        self.cur_mem_file.seek(0)
        cfg = self.default_config.copy()
        cfg["delay"] = 1.1
        loop = monitor.Monitor(DummySleepPlugin, cfg)
        with self.assertRaises(monitor.NoNetworkResultError) as e:
            loop.run_once(self.basic_input_event)
        job_oom_result = e.exception.result

        # standardise variable timestamps
        print(job_oom_result.state)
        self.assertEqual(
            job_oom_result.state,
            State(
                State.Label.ERROR_OOM,
                failure_name="Out of Memory",
                # At 90 rather than 100 because the refinement lowers the memory limit
                message="Plugin DummySleepPlugin failed to complete job '1234' because it ran out of memory, memory limit is 100B and memory usage was 90B which is 90.0% memory usage.",
            ),
        )

    @pytest.mark.timeout(7)
    def test_oom_failure_multiplugin(self):
        """Tests that the plugin exits with an exception if memory usage is over the error threshold."""
        self.cur_mem_file.write("0")
        self.cur_mem_file.flush()
        self.cur_mem_file.seek(0)
        config = self.default_config.copy()
        config["memory_file_path"] = self.cur_mem_file.name
        config["server"] = self.server + "/null"
        loop = monitor.Monitor(SleepMultiPlugin, config)
        loop.run_loop(1)

        r = httpx.get("%s/mock/get_var/all_requests" % self.server)
        # Get the OOM event which should occur after the two plugin registration events.
        out_evt: azm.StatusEvent = azm.StatusEvent(**r.json()[2][0])

        # standardise variable timestamps
        self.assertEqual(out_evt.entity.status, State.Label.ERROR_OOM.value)
        self.assertEqual(out_evt.entity.error, "Out of Memory")
        print(f"Actual Message: {out_evt.entity.message}")
        self.assertEqual(
            out_evt.entity.message,
            "Plugin SleepMultiPlugin-MultiPluginName failed to complete job '1234' because it ran out of memory, memory limit is 100B and memory usage was 90B which is 90.0% memory usage.",
        )
        self.assertEqual(
            out_evt.author,
            azm.Author(category="plugin", name="SleepMultiPlugin-MultiPluginName", version="1.0-1.0", security=None),
        )

    @pytest.mark.timeout(20)
    def test_high_memory_usage_warnings(self):
        """Tests that warnings are raised if the plugin has high memory usage but not error level usage."""
        self.cur_mem_file.write("85")
        self.cur_mem_file.flush()
        self.cur_mem_file.seek(0)
        cfg = self.default_config.copy()
        cfg["delay"] = 1.1
        loop = monitor.Monitor(DummySleepPlugin, cfg)
        with self.assertLogs() as caught_logs:
            loop.run_once(self.basic_input_event)

        found = False
        for log_list in caught_logs:
            for log_record in log_list:
                print(log_record)
                if "'1234' is nearly out of memory it is at 85.0% memory usage" in str(log_record):
                    found = True
        self.assertTrue(found, "Failed to find warning that runner is at 85% memory usage.")

    @pytest.mark.timeout(20)
    def test_high_memory_usage_after_inactivate_memory(self):
        """Tests that warnings are raised if the plugin has high memory usage but not error level usage."""
        self.cur_mem_file.write("95")
        self.cur_mem_file.flush()
        self.cur_mem_file.seek(0)
        cfg = self.default_config.copy()
        cfg["delay"] = 1.1
        loop = monitor.Monitor(DummySleepPlugin, cfg)
        with self.assertLogs() as caught_logs:
            loop.run_once(self.basic_input_event)

        found = False
        for log_list in caught_logs:
            for log_record in log_list:
                print(log_record)
                if "'1234' is nearly out of memory it is at 85.0% memory usage" in str(log_record):
                    found = True
        self.assertTrue(found, "Failed to find warning that runner is at 85% memory usage.")

    @pytest.mark.timeout(4)
    def test_setup_fails_and_disables_mem_tracking(self):
        """Tests that setup errors when trying to read cgroup memory."""
        self.max_mem_file.write("MAXMEM")
        self.max_mem_file.flush()
        self.max_mem_file.seek(0)
        loop = monitor.Monitor(DummySleepPlugin, self.default_config.copy())
        result = loop.run_once(self.basic_input_event)

        # Ensure a result is acquired rather than an OOM error
        print(result[None])
        self.assertEqual(len(result), 1)
        self.assertEqual(result[None].state, State(State.Label.COMPLETED_EMPTY))

    @pytest.mark.timeout(20)
    def test_kill_children_called_oom(self):
        """Test to see if child processes are killed on out of memory"""
        cfg = self.default_config.copy()
        cfg["delay"] = 1.1
        loop = monitor.Monitor(DummySleepPlugin, cfg)

        called = False

        def raise_exception_instead(*args, **kwargs):
            nonlocal called
            called = True
            raise CustomTestException()

        loop._kill_child_processes = raise_exception_instead
        try:
            loop.run_once(self.basic_input_event)
        except CustomTestException:
            pass
        self.assertTrue(
            called, "kill_child_process wasn't called even though out of memory should have been detected."
        )

        # Cleanup leftover -result file this test leaves behind.
        for f in os.listdir(tempfile.gettempdir()):
            out_path = os.path.join(tempfile.gettempdir(), f)
            if os.path.isfile(out_path) and f.endswith(monitor.LOCAL_RESULT_SUFFIX):
                os.remove(out_path)



from __future__ import annotations

import contextlib
import ctypes
import multiprocessing
import os
import subprocess
import tempfile
import time
import unittest
import warnings
from multiprocessing import Process
from queue import Empty
from typing import Any, ClassVar

import httpx
import pytest

from azul_runner import coordinator, settings

from . import mock_dispatcher as md
from . import plugin_support as sup


def modify_file_in_background_causing_crash(filepath: str, queue: multiprocessing.Queue):
    # wait one second while emptying the queue and then modify the test file and exit.
    start_time = time.time()
    while start_time + 1 > time.time():
        # NOTE - without continually emptying the queue otherwise the plugin can get stuck putting elements into the
        # multiprocessing queue.
        # If there is no timeout on a multiprocessing queue put statement the application will un-expectedly hang.
        with contextlib.suppress(Empty):
            queue.get(block=False)
    with open(os.path.join(filepath, "tmp.txt"), "w") as f:
        f.write("2")


class TestPluginExecutionWrapper(unittest.TestCase):
    """
    Tests the handling of plugin execution by Plugin._exec_wrapper, using the sup.DummyPlugin class and TestPlugin template.
    """

    PLUGIN_TO_TEST = sup.DummyPlugin

    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server, suitable for passing to a plugin's config['server']
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()

        # Dummy shared memory queue and ctype
        cls.dummy_queue: multiprocessing.Queue = multiprocessing.Queue()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()
        cls.dummy_queue.close()

    def _inner_test_watch(self, filepath: str, watch_type: str = None):
        """Common code between normal and git watchers."""
        config_dict = {"server": self.server + "/test_data", "watch_path": filepath, "watch_wait": 0}
        if watch_type:
            config_dict["watch_type"] = watch_type
        loop = coordinator.Coordinator(sup.DummyPlugin, settings.parse_config(sup.DummyPlugin, config_dict))

        p = Process(
            target=modify_file_in_background_causing_crash,
            args=(
                filepath,
                self.dummy_queue,
            ),
        )
        p.start()

        with self.assertRaises(coordinator.RecreateException):
            loop.run_loop(queue=self.dummy_queue, job_limit=None)
        # kill the watchdog threads to reduce chance of weird stuff during testing
        del loop

    @pytest.mark.timeout(10)
    def test_watch(self):
        """Tests that coordinator raises a recreation exception when a change occurs."""
        with tempfile.TemporaryDirectory() as filepath:
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("1")
            self._inner_test_watch(filepath)

    @pytest.mark.timeout(10)
    def test_watch_git(self):
        """Tests git watch."""

        with tempfile.TemporaryDirectory() as filepath:
            subprocess.call(["git", "init"], cwd=filepath)
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("1")
            subprocess.call(["git", "add", "tmp.txt"], cwd=filepath)
            subprocess.call(["git", "commit", "-m", '"add file"'], cwd=filepath)

            self._inner_test_watch(filepath, watch_type="git")

    class DPWatchGitMissing(sup.DummyPlugin):
        def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
            super().__init__(config)
            with open(os.path.join(config["test_input_data"]["filepath"], "tmp.txt"), "r") as f:
                self.retval = f.read()

        def execute(self, job):
            self.add_feature_values("example_string", self.retval)

    def test_watch_git_missing(self):
        with tempfile.TemporaryDirectory() as filepath:
            self.assertRaisesRegex(
                coordinator.CriticalError,
                r"is git not installed or .* not a valid git checkout",
                coordinator.Coordinator,
                *(
                    self.DPWatchGitMissing,
                    settings.parse_config(
                        self.DPWatchGitMissing,
                        {
                            "watch_path": filepath,
                            "watch_wait": 0,
                            "watch_type": "git",
                            "test_input_data": {"filepath": filepath},
                        },
                    ),
                ),
            )

    class DPTestNoWatch(sup.DummyPlugin):
        def execute(self, job):
            # Sleep up to job limit (6) waiting up to 3 seconds for a Recreate error to be generated.
            time.sleep(0.5)

    @pytest.mark.timeout(10)
    def test_no_watch(self):
        """Test that no exception is raised when files are modified that would cause watch to trigger if configured."""

        with tempfile.TemporaryDirectory() as filepath:
            with open(os.path.join(filepath, "tmp.txt"), "w") as f:
                f.write("1")

            loop = coordinator.Coordinator(
                self.DPTestNoWatch, settings.parse_config(self.DPTestNoWatch, {"server": self.server + "/test_data"})
            )

            p = Process(
                target=modify_file_in_background_causing_crash,
                args=(
                    filepath,
                    self.dummy_queue,
                ),
            )
            p.start()
            with self.assertRaises(SystemExit):
                loop.run_loop(queue=self.dummy_queue, job_limit=6)



"""Example plugins used in readme."""

from azul_runner import (
    BinaryPlugin,
    Feature,
    FeatureType,
    Job,
    State,
    add_settings,
    cmdline_run,
)


class LookForThings(BinaryPlugin):
    """Look for things."""

    VERSION = "1.0"
    FEATURES = [
        Feature("tag", "Custom tag", FeatureType.String),
    ]
    SETTINGS = add_settings(
        filter_data_types={"content": ["text/plain"]},
    )

    def execute(self, job: Job):
        """Find peanuts."""
        data = job.get_data()
        # 'data' is a file-like object that supports seeking and being read from
        # (The content may be retrieved in parts if the file is large and non-local)
        header: bytes = data.read(7)
        if header == b"PEANUT:":
            # create a tag
            self.add_feature_values("tag", "may contain nuts")
            # add the next 24 bytes as a child
            c = self.add_child_with_data(
                relationship={"label": "peanut"},
                data=data.read(24),
            )
            c.add_feature_values("tag", "may be hard to crack")
        else:
            return State.Label.OPT_OUT


if __name__ == "__main__":
    cmdline_run(plugin=LookForThings)


class LookForMZ1(BinaryPlugin):
    """Look for MZ type 1."""

    VERSION = "1.0"
    FEATURES = [
        Feature("tag", "Tags files that might be .EXEs", FeatureType.String),
    ]

    def execute(self, job: Job):
        """Find MZ."""
        data = job.get_data()
        # 'data' is a file-like object that supports seeking and being read from
        # (The content may be retrieved in parts if the file is large and non-local)
        header: bytes = data.read(2)
        if header == b"MZ":
            self.add_feature_values("tag", "might be EXE")
        else:
            return State.Label.OPT_OUT


class LookForMZ2(BinaryPlugin):
    """Look for MZ type 2."""

    VERSION = "1.0"
    FEATURES = [
        Feature("tag", "Tags files that might be .EXEs", FeatureType.String),
    ]
    SETTINGS = add_settings(
        filter_data_types={"content": ["executable/windows/pe32", "executable/windows/dos"]},
    )

    def execute(self, job: Job):
        """Find MZ."""
        # This plugin will only run over files that have been identified
        #  by azul-runner/dispatcher as EXE files.
        self.add_feature_values("tag", "It's an EXE!")


class LookForMZ3(BinaryPlugin):
    """Look for MZ type 3."""

    VERSION = "1.0"
    FEATURES = [
        Feature("tag", "Tags files that might be .EXEs", FeatureType.String),
    ]
    SETTINGS = add_settings(
        filter_data_types={"content": ["executable/windows/pe32", "executable/windows/dos"]},
    )

    def execute(self, job: Job):
        """Find MZ."""
        buffer = job.get_data().read(64)
        # add child binary
        c = self.add_child_with_data(
            relationship={"label": "First 64 bytes"},
            data=buffer,
        )
        # add feature to the child binary
        c.add_feature_values("tag", "the extracted child")
        # add feature to the original incoming binary
        self.add_feature_values("tag", ["Might be an exe", "Extracted header"])


if __name__ == "__main__":
    cmdline_run(plugin=LookForThings)



import datetime
import hashlib
import json
import unittest
from unittest import mock

from azul_bedrock import models_network as azm

from azul_runner import (
    FV,
    Event,
    EventData,
    EventParent,
    Feature,
    JobResult,
    State,
    network_transform,
)

from . import plugin_support as sup


def dump(x):
    return json.loads(x.model_dump_json(exclude_defaults=True))


def gen_api_content(x, label: azm.DataLabel = azm.DataLabel.CONTENT):
    return azm.Datastream(
        identify_version=1,
        label=label,
        size=5,
        md5=x,
        sha1=x,
        sha256=x,
        sha512=x,
        mime="mime",
        magic="magic",
        file_format="file/type",
        file_extension="ft",
        ssdeep="12:some:thing",
        tlsh="T1xxxxxxx",
    )


def gen_src_event():
    timestamp = datetime.datetime(year=2000, month=1, day=1, tzinfo=datetime.timezone.utc)
    return azm.BinaryEvent(
        model_version=azm.CURRENT_MODEL_VERSION,
        kafka_key="dummy-id",
        dequeued="dummy-dequeued-id",
        action=azm.BinaryAction.Extracted,
        timestamp=timestamp,
        source=azm.Source(
            name="sauce",
            path=[
                azm.PathNode(
                    action=azm.BinaryAction.Extracted,
                    timestamp=timestamp,
                    author=azm.Author(category="plugin", name="system", version="1", security=None),
                    sha256="hash",
                )
            ],
            timestamp=timestamp,
            security=None,
            references={"keen": "eye"},
        ),
        author=azm.Author(category="plugin", name="system", version="1", security=None),
        entity=azm.BinaryEvent.Entity(
            sha256="hash",
            datastreams=[gen_api_content("hash"), gen_api_content("hash5", azm.DataLabel.TEXT)],
        ),
    )


class TestGenEvents(unittest.TestCase):
    maxDiff = None

    class P(sup.DummyPluginMinimum):
        FEATURES = [
            Feature(name="feature1", desc="", type=azm.FeatureType.Integer),
            Feature(name="feature2", desc="", type=azm.FeatureType.String),
        ]

    p = P()

    def test_gen_event_enriched(self):
        src = gen_src_event()
        posted = {"hash2": gen_api_content("hash2", azm.DataLabel.TEXT)}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256=src.entity.sha256,
            info={"test": "data"},
            features={"feature1": [FV(1), FV(2)], "feature2": [FV("1"), FV("2")]},
            data=[],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        res = dump(network_transform._gen_event_output(self.p, posted, author, src, event, now))
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "source": {
                    "references": {"keen": "eye"},
                    "path": [
                        {
                            "sha256": "hash",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                            "action": "extracted",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                        },
                        {
                            "sha256": "hash",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "action": "enriched",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                        },
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "name": "sauce",
                },
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "action": "enriched",
                "timestamp": "2010-01-01T00:00:00+00:00",
                "entity": {
                    "info": {"test": "data"},
                    "features": [
                        {"type": "integer", "name": "feature1", "value": "1"},
                        {"type": "integer", "name": "feature1", "value": "2"},
                        {"type": "string", "name": "feature2", "value": "1"},
                        {"type": "string", "name": "feature2", "value": "2"},
                    ],
                    "sha256": "hash",
                },
            },
        )

    def test_using_pusher_generation(self):
        """Test generating an event using the pusher event generation."""
        src = gen_src_event()
        posted = {"hash2": gen_api_content("hash2", azm.DataLabel.TEXT)}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256=src.entity.sha256,
            info={"test": "data"},
            features={"feature1": [FV(1), FV(2)], "feature2": [FV("1"), FV("2")]},
            data=[],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        custom_p = self.P()
        custom_p._IS_USING_PUSHER = True
        res = dump(network_transform._gen_event_output(custom_p, posted, author, src, event, now))
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "timestamp": "2010-01-01T00:00:00+00:00",
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "entity": {
                    "sha256": "hash",
                    "features": [
                        {"name": "feature1", "type": "integer", "value": "1"},
                        {"name": "feature1", "type": "integer", "value": "2"},
                        {"name": "feature2", "type": "string", "value": "1"},
                        {"name": "feature2", "type": "string", "value": "2"},
                    ],
                    "datastreams": [
                        {
                            "sha256": "hash",
                            "sha512": "hash",
                            "sha1": "hash",
                            "md5": "hash",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "content",
                        },
                        {
                            "sha256": "hash5",
                            "sha512": "hash5",
                            "sha1": "hash5",
                            "md5": "hash5",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "text",
                        },
                    ],
                    "info": {"test": "data"},
                },
                "action": "extracted",
                "source": {
                    "name": "sauce",
                    "path": [
                        {
                            "sha256": "hash",
                            "action": "extracted",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                        }
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "references": {"keen": "eye"},
                },
            },
        )

    def test_using_pusher_generation_both_events_have_features(self):
        """Test generating an event using the pusher event generation, where the source event has features to."""
        src = gen_src_event()
        src.action = azm.BinaryAction.Mapped
        src.source.path[0].action = azm.BinaryAction.Mapped
        src.entity.features.append(
            azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="SPECIAL_SOURCE_FILENAME!")
        )
        posted = {"hash2": gen_api_content("hash2", azm.DataLabel.TEXT)}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256=src.entity.sha256,
            info={"test": "data"},
            features={"feature1": [FV(1), FV(2)], "feature2": [FV("1"), FV("2")]},
            data=[],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        custom_p = self.P()
        custom_p._IS_USING_PUSHER = True
        res = dump(network_transform._gen_event_output(custom_p, posted, author, src, event, now))
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "timestamp": "2010-01-01T00:00:00+00:00",
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "entity": {
                    "sha256": "hash",
                    "features": [
                        {"name": "feature1", "type": "integer", "value": "1"},
                        {"name": "feature1", "type": "integer", "value": "2"},
                        {"name": "feature2", "type": "string", "value": "1"},
                        {"name": "feature2", "type": "string", "value": "2"},
                        {"name": "filename", "type": "filepath", "value": "SPECIAL_SOURCE_FILENAME!"},
                    ],
                    "datastreams": [
                        {
                            "sha256": "hash",
                            "sha512": "hash",
                            "sha1": "hash",
                            "md5": "hash",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "content",
                        },
                        {
                            "sha256": "hash5",
                            "sha512": "hash5",
                            "sha1": "hash5",
                            "md5": "hash5",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "text",
                        },
                    ],
                    "info": {"test": "data"},
                },
                "action": "mapped",
                "source": {
                    "name": "sauce",
                    "path": [
                        {
                            "sha256": "hash",
                            "action": "mapped",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                        }
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "references": {"keen": "eye"},
                },
            },
        )

    def test_using_pusher_generation_only_source_event_has_features(self):
        """Test generating an event using the pusher event generation,
        where the source event has features and the child event doesn't.

        This accounts for an edge case where the parent features could be duplicated.
        """
        src = gen_src_event()
        src.action = azm.BinaryAction.Sourced
        src.source.path[0].action = azm.BinaryAction.Sourced
        src.entity.features.append(
            azm.FeatureValue(name="filename", type=azm.FeatureType.Filepath, value="SPECIAL_SOURCE_FILENAME!")
        )
        posted = {"hash2": gen_api_content("hash2", azm.DataLabel.TEXT)}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256=src.entity.sha256,
            info={"test": "data"},
            data=[],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        custom_p = self.P()
        custom_p._IS_USING_PUSHER = True
        res = dump(network_transform._gen_event_output(custom_p, posted, author, src, event, now))
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "timestamp": "2010-01-01T00:00:00+00:00",
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "entity": {
                    "sha256": "hash",
                    "features": [{"name": "filename", "type": "filepath", "value": "SPECIAL_SOURCE_FILENAME!"}],
                    "datastreams": [
                        {
                            "sha256": "hash",
                            "sha512": "hash",
                            "sha1": "hash",
                            "md5": "hash",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "content",
                        },
                        {
                            "sha256": "hash5",
                            "sha512": "hash5",
                            "sha1": "hash5",
                            "md5": "hash5",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "text",
                        },
                    ],
                    "info": {"test": "data"},
                },
                "action": "sourced",
                "source": {
                    "name": "sauce",
                    "path": [
                        {
                            "sha256": "hash",
                            "action": "sourced",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                        }
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "references": {"keen": "eye"},
                },
            },
        )

    def test_gen_event_augmented(self):
        src = gen_src_event()
        posted = {"hash2": gen_api_content("hash2", azm.DataLabel.TEXT)}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256=src.entity.sha256,
            info={"test": "data"},
            features={"feature1": [FV(1), FV(2)], "feature2": [FV("1"), FV("2")]},
            data=[EventData(hash="hash2", label=azm.DataLabel.TEXT, language="javascript")],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        res = dump(network_transform._gen_event_output(self.p, posted, author, src, event, now))
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "source": {
                    "references": {"keen": "eye"},
                    "path": [
                        {
                            "sha256": "hash",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                            "action": "extracted",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                        },
                        {
                            "sha256": "hash",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "action": "augmented",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                        },
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "name": "sauce",
                },
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "action": "augmented",
                "timestamp": "2010-01-01T00:00:00+00:00",
                "entity": {
                    "datastreams": [
                        {
                            "identify_version": 1,
                            "label": "content",
                            "size": 5,
                            "sha1": "hash",
                            "sha256": "hash",
                            "md5": "hash",
                            "magic": "magic",
                            "mime": "mime",
                            "sha512": "hash",
                            "file_extension": "ft",
                            "file_format": "file/type",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                        },
                        {
                            "identify_version": 1,
                            "language": "javascript",
                            "label": "text",
                            "size": 5,
                            "sha1": "hash2",
                            "sha256": "hash2",
                            "md5": "hash2",
                            "magic": "magic",
                            "mime": "mime",
                            "sha512": "hash2",
                            "file_extension": "ft",
                            "file_format": "file/type",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                        },
                    ],
                    "info": {"test": "data"},
                    "features": [
                        {"type": "integer", "name": "feature1", "value": "1"},
                        {"type": "integer", "name": "feature1", "value": "2"},
                        {"type": "string", "name": "feature2", "value": "1"},
                        {"type": "string", "name": "feature2", "value": "2"},
                    ],
                    "sha256": "hash",
                },
            },
        )

    def test_gen_event_extracted_1deep(self):
        src = gen_src_event()
        posted = {"hash2": gen_api_content("hash2", azm.DataLabel.TEXT), "hash3": gen_api_content("hash3")}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256="hash3",
            parent=EventParent(
                sha256=src.entity.sha256,
            ),  # parent is top level binary
            relationship={"decoded": "rot26"},
            info={"test": "data"},
            features={
                "feature1": [FV(1), FV(2)],
                "feature2": [FV("1"), FV("2")],
                "filename": [FV("a.exe"), FV("b.exe")],
            },
            data=[
                EventData(hash="hash3", label=azm.DataLabel.CONTENT),
                EventData(hash="hash2", label=azm.DataLabel.TEXT, language="javascript"),
            ],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        res = dump(network_transform._gen_event_extracted(self.p, posted, author, src, event, now))
        print(res)
        self.assertEqual(
            res,
            {
                "kafka_key": "runner-placeholder",
                "action": "extracted",
                "model_version": azm.CURRENT_MODEL_VERSION,
                "timestamp": "2010-01-01T00:00:00+00:00",
                "source": {
                    "name": "sauce",
                    "path": [
                        {
                            "sha256": "hash",
                            "action": "extracted",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                        },
                        {
                            "sha256": "hash3",
                            "action": "extracted",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "relationship": {"decoded": "rot26"},
                            "file_format": "file/type",
                            "size": 5,
                            "filename": "a.exe",
                        },
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "references": {"keen": "eye"},
                },
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "entity": {
                    "sha256": "hash3",
                    "sha512": "hash3",
                    "sha1": "hash3",
                    "md5": "hash3",
                    "ssdeep": "12:some:thing",
                    "tlsh": "T1xxxxxxx",
                    "size": 5,
                    "file_format": "file/type",
                    "file_extension": "ft",
                    "mime": "mime",
                    "magic": "magic",
                    "info": {"test": "data"},
                    "features": [
                        {"name": "file_format", "type": "string", "value": "file/type"},
                        {"name": "file_extension", "type": "string", "value": "ft"},
                        {"name": "magic", "type": "string", "value": "magic"},
                        {"name": "mime", "type": "string", "value": "mime"},
                        {"name": "feature1", "type": "integer", "value": "1"},
                        {"name": "feature1", "type": "integer", "value": "2"},
                        {"name": "feature2", "type": "string", "value": "1"},
                        {"name": "feature2", "type": "string", "value": "2"},
                        {"name": "filename", "type": "filepath", "value": "a.exe"},
                        {"name": "filename", "type": "filepath", "value": "b.exe"},
                    ],
                    "datastreams": [
                        {
                            "sha256": "hash2",
                            "sha512": "hash2",
                            "sha1": "hash2",
                            "md5": "hash2",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "text",
                            "language": "javascript",
                        },
                        {
                            "sha256": "hash3",
                            "sha512": "hash3",
                            "sha1": "hash3",
                            "md5": "hash3",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "content",
                        },
                    ],
                },
            },
        )

    def test_gen_event_extracted_4deep(self):
        src = gen_src_event()
        posted = {
            "hash2": gen_api_content("hash2", azm.DataLabel.TEXT),
            "hash3": gen_api_content("hash3"),
            "parent1": gen_api_content("parent1"),
            "parent2": gen_api_content("parent2"),
        }
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256="hash3",
            parent=EventParent(
                sha256="parent1",
                filename="double.exe",
                parent=EventParent(
                    sha256="parent2",
                    filename="trouble.exe",
                    parent=EventParent(
                        sha256=src.entity.sha256,
                    ),
                ),
            ),  # parent is top level binary
            relationship={"decoded": "rot26"},
            info={"test": "data"},
            features={
                "feature1": [FV(1), FV(2)],
                "feature2": [FV("1"), FV("2")],
                "filename": [FV("a.exe"), FV("b.exe")],
            },
            data=[
                EventData(hash="hash3", label=azm.DataLabel.CONTENT),
                EventData(hash="hash2", label=azm.DataLabel.TEXT, language="javascript"),
            ],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        res = dump(network_transform._gen_event_extracted(self.p, posted, author, src, event, now))
        print(res)
        self.assertEqual(
            res,
            {
                "kafka_key": "runner-placeholder",
                "action": "extracted",
                "model_version": azm.CURRENT_MODEL_VERSION,
                "timestamp": "2010-01-01T00:00:00+00:00",
                "source": {
                    "name": "sauce",
                    "path": [
                        {
                            "sha256": "hash",
                            "action": "extracted",
                            "timestamp": "2000-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "system", "version": "1"},
                        },
                        {
                            "sha256": "parent2",
                            "action": "extracted",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "file_format": "file/type",
                            "size": 5,
                            "filename": "trouble.exe",
                        },
                        {
                            "sha256": "parent1",
                            "action": "extracted",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "file_format": "file/type",
                            "size": 5,
                            "filename": "double.exe",
                        },
                        {
                            "sha256": "hash3",
                            "action": "extracted",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "relationship": {"decoded": "rot26"},
                            "file_format": "file/type",
                            "size": 5,
                            "filename": "a.exe",
                        },
                    ],
                    "timestamp": "2000-01-01T00:00:00+00:00",
                    "references": {"keen": "eye"},
                },
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "entity": {
                    "sha256": "hash3",
                    "sha512": "hash3",
                    "sha1": "hash3",
                    "md5": "hash3",
                    "ssdeep": "12:some:thing",
                    "tlsh": "T1xxxxxxx",
                    "size": 5,
                    "file_format": "file/type",
                    "file_extension": "ft",
                    "mime": "mime",
                    "magic": "magic",
                    "info": {"test": "data"},
                    "features": [
                        {"name": "file_format", "type": "string", "value": "file/type"},
                        {"name": "file_extension", "type": "string", "value": "ft"},
                        {"name": "magic", "type": "string", "value": "magic"},
                        {"name": "mime", "type": "string", "value": "mime"},
                        {"name": "feature1", "type": "integer", "value": "1"},
                        {"name": "feature1", "type": "integer", "value": "2"},
                        {"name": "feature2", "type": "string", "value": "1"},
                        {"name": "feature2", "type": "string", "value": "2"},
                        {"name": "filename", "type": "filepath", "value": "a.exe"},
                        {"name": "filename", "type": "filepath", "value": "b.exe"},
                    ],
                    "datastreams": [
                        {
                            "sha256": "hash2",
                            "sha512": "hash2",
                            "sha1": "hash2",
                            "md5": "hash2",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "text",
                            "language": "javascript",
                        },
                        {
                            "sha256": "hash3",
                            "sha512": "hash3",
                            "sha1": "hash3",
                            "md5": "hash3",
                            "ssdeep": "12:some:thing",
                            "tlsh": "T1xxxxxxx",
                            "size": 5,
                            "file_format": "file/type",
                            "file_extension": "ft",
                            "mime": "mime",
                            "magic": "magic",
                            "identify_version": 1,
                            "label": "content",
                        },
                    ],
                },
            },
        )

    def test_gen_processing_events_enriched(self):
        src = gen_src_event()
        posted = {
            "hash2": gen_api_content("hash2", azm.DataLabel.TEXT),
            "hash3": gen_api_content("hash3"),
            "parent1": gen_api_content("parent1"),
            "parent2": gen_api_content("parent2"),
        }
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(sha256=src.entity.sha256, info={"test": "data"})
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        result = JobResult(state=State(), events=[event], runtime=5)
        with mock.patch("azul_runner.network_transform._get_now", lambda: now):
            res = network_transform.gen_processing_events(self.p, posted, author, src, result)
        res = dump(res)
        res["entity"].pop("input")  # dont care about this
        print(res)
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "author": {"category": "plugin", "version": "1", "name": "generic"},
                "entity": {
                    "status": "completed",
                    "results": [
                        {
                            "model_version": azm.CURRENT_MODEL_VERSION,
                            "kafka_key": "runner-placeholder",
                            "source": {
                                "path": [
                                    {
                                        "author": {"category": "plugin", "version": "1", "name": "system"},
                                        "sha256": "hash",
                                        "action": "extracted",
                                        "timestamp": "2000-01-01T00:00:00+00:00",
                                    },
                                    {
                                        "author": {"category": "plugin", "version": "1", "name": "generic"},
                                        "sha256": "hash",
                                        "action": "enriched",
                                        "timestamp": "2010-01-01T00:00:00+00:00",
                                    },
                                ],
                                "references": {"keen": "eye"},
                                "name": "sauce",
                                "timestamp": "2000-01-01T00:00:00+00:00",
                            },
                            "author": {"category": "plugin", "version": "1", "name": "generic"},
                            "action": "enriched",
                            "entity": {
                                "sha256": "hash",
                                "info": {"test": "data"},
                            },
                            "timestamp": "2010-01-01T00:00:00+00:00",
                        }
                    ],
                    "runtime": 5.0,
                },
                "timestamp": "2010-01-01T00:00:00+00:00",
            },
        )

    def test_gen_processing_events_extracted(self):
        src = gen_src_event()
        posted = {
            "hash2": gen_api_content("hash2", azm.DataLabel.TEXT),
            "hash3": gen_api_content("hash3"),
            "parent1": gen_api_content("parent1"),
            "parent2": gen_api_content("parent2"),
        }
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256="hash3",
            parent=EventParent(sha256=src.entity.sha256),  # parent is top level binary
            relationship={"decoded": "rot26"},
            info={"test": "data"},
            data=[
                EventData(hash="hash3", label=azm.DataLabel.CONTENT),
            ],
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        result = JobResult(state=State(), events=[event], runtime=5)
        with mock.patch("azul_runner.network_transform._get_now", lambda: now):
            res = network_transform.gen_processing_events(self.p, posted, author, src, result)
        res = dump(res)
        res["entity"].pop("input")  # dont care about this
        print(res)
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "author": {"version": "1", "category": "plugin", "name": "generic"},
                "entity": {
                    "results": [
                        {
                            "author": {"version": "1", "category": "plugin", "name": "generic"},
                            "entity": {
                                "datastreams": [
                                    {
                                        "identify_version": 1,
                                        "sha256": "hash3",
                                        "mime": "mime",
                                        "sha512": "hash3",
                                        "sha1": "hash3",
                                        "size": 5,
                                        "magic": "magic",
                                        "label": "content",
                                        "md5": "hash3",
                                        "file_extension": "ft",
                                        "file_format": "file/type",
                                        "ssdeep": "12:some:thing",
                                        "tlsh": "T1xxxxxxx",
                                    }
                                ],
                                "sha256": "hash3",
                                "info": {"test": "data"},
                                "features": [
                                    {"type": "string", "value": "mime", "name": "mime"},
                                    {"type": "string", "value": "magic", "name": "magic"},
                                ],
                                "sha512": "hash3",
                                "sha1": "hash3",
                                "size": 5,
                                "md5": "hash3",
                                "file_extension": "ft",
                                "file_format": "file/type",
                                "ssdeep": "12:some:thing",
                                "tlsh": "T1xxxxxxx",
                            },
                            "action": "extracted",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                            "model_version": azm.CURRENT_MODEL_VERSION,
                            "kafka_key": "",
                            "source": {
                                "timestamp": "2000-01-01T00:00:00+00:00",
                                "name": "sauce",
                                "path": [
                                    {
                                        "author": {"version": "1", "category": "plugin", "name": "system"},
                                        "sha256": "hash",
                                        "action": "extracted",
                                        "timestamp": "2000-01-01T00:00:00+00:00",
                                    },
                                    {
                                        "author": {"version": "1", "category": "plugin", "name": "generic"},
                                        "sha256": "hash3",
                                        "action": "extracted",
                                        "timestamp": "2010-01-01T00:00:00+00:00",
                                        "size": 5,
                                        "relationship": {"decoded": "rot26"},
                                    },
                                ],
                                "references": {"keen": "eye"},
                            },
                        }
                    ],
                    "status": "completed",
                    "runtime": 5.0,
                    "kafka_key": "-generic-1",
                },
                "timestamp": "2010-01-01T00:00:00+00:00",
                "kafka_key": "",
            },
        )

    def test_gen_processing_events_extracted(self):
        src = gen_src_event()
        posted = {}
        author = azm.Author(category="plugin", name="generic", version="1", security=None)
        event = Event(
            sha256="hash3",
            parent=EventParent(sha256=src.entity.sha256),  # parent is top level binary
            relationship={"decoded": "rot26"},
            info={"test": "data"},
            data=[
                EventData(hash="hash3", label=azm.DataLabel.CONTENT),
            ],
            md5="md5",
            sha512="sha512",
            size=5454,
        )
        now = datetime.datetime(year=2010, month=1, day=1, tzinfo=datetime.timezone.utc)
        result = JobResult(state=State(), events=[event], runtime=5)
        with mock.patch("azul_runner.network_transform._get_now", lambda: now):
            res = network_transform.gen_processing_events(self.p, posted, author, src, result)
        res = dump(res)
        res["entity"].pop("input")  # dont care about this
        print(res)
        self.assertEqual(
            res,
            {
                "model_version": azm.CURRENT_MODEL_VERSION,
                "kafka_key": "runner-placeholder",
                "timestamp": "2010-01-01T00:00:00+00:00",
                "author": {"category": "plugin", "name": "generic", "version": "1"},
                "entity": {
                    "status": "completed",
                    "runtime": 5.0,
                    "results": [
                        {
                            "model_version": azm.CURRENT_MODEL_VERSION,
                            "kafka_key": "runner-placeholder",
                            "timestamp": "2010-01-01T00:00:00+00:00",
                            "author": {"category": "plugin", "name": "generic", "version": "1"},
                            "entity": {
                                "sha256": "hash3",
                                "sha512": "sha512",
                                "md5": "md5",
                                "size": 5454,
                                "info": {"test": "data"},
                            },
                            "action": "extracted",
                            "source": {
                                "name": "sauce",
                                "timestamp": "2000-01-01T00:00:00+00:00",
                                "references": {"keen": "eye"},
                                "path": [
                                    {
                                        "sha256": "hash",
                                        "action": "extracted",
                                        "timestamp": "2000-01-01T00:00:00+00:00",
                                        "author": {"category": "plugin", "name": "system", "version": "1"},
                                    },
                                    {
                                        "sha256": "hash3",
                                        "action": "extracted",
                                        "timestamp": "2010-01-01T00:00:00+00:00",
                                        "author": {"category": "plugin", "name": "generic", "version": "1"},
                                        "relationship": {"decoded": "rot26"},
                                        "size": 5454,
                                    },
                                ],
                            },
                        }
                    ],
                },
            },
        )
